# -*- coding: utf-8 -*-
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from DateTime import DateTime
from appy import Object as O
from appy.px import Px
from appy.gen import utils as gutils
from appy.gen.indexer import defaultIndexes
from appy.gen.navigate import Batch
from appy.shared import utils as sutils
from group import Group

# Error messages ---------------------------------------------------------------
WRONG_FIELD = 'Field "%s" does not exist on %s'

# ------------------------------------------------------------------------------
class Gridder:
    '''Specification about how to produce search results in grid mode. An
       instance of Gridder can be specified in static attribute
       SomeAppyClass.gridder.'''

    def __init__(self, width='350px', gap='20px', justifyContent='space-evenly',
                 alignContent='space-evenly'):
        # The minimum width of every grid element
        self.width = width
        # The gap between grid elements. Will be passed to CSS property
        # "grid-gap".
        self.gap = gap
        # Specify how to align the whole grid horizontally inside the container.
        # Will be passed to CSS property "justify-content".
        self.justifyContent = justifyContent
        # Specify how to align the whole grid vertically inside the container.
        # Will be passed to CSS property "align-content".
        self.alignContent = alignContent

    def getContainerStyle(self):
        '''Returns the CSS styles that must be applied to the grid container
           element.'''
        return 'display: grid; ' \
               'grid-template-columns: repeat(auto-fill, minmax(%s, 1fr)); ' \
               'grid-gap: %s; justify-content: %s; align-content: %s' % \
               (self.width, self.gap, self.justifyContent, self.alignContent)

# ------------------------------------------------------------------------------
class Mode:
    '''Search results can be displayed in various "modes": as a list, grid,
       calendar, etc. This class is the abstract base class for such modes.
       A concrete Mode instance is created very time search results must be
       computed.'''
    # The default mode(s) for displaying instances of any Appy class
    default = ('list',)
    # All available predefined concrete modes
    concrete = ('list', 'grid', 'calendar')

    # The list of custom actions that can be triggered on search results
    pxActions = Px('''
     <table>
      <tr><td for="action in actions"
            var2="field=action; fieldName=field.name;
                  multi=True">:action.pxRender</td></tr>
     </table>''')

    @staticmethod
    def get(klass, className, uiSearch, inPopup, tool, req, dir):
        '''Create and return the Mode instance corresponding to the current
           result mode to apply to a list of p_klass instances.'''
        name = req.get('resultMode') or uiSearch.getModes(klass, tool)[0]
        # Determine the concrete mode class
        if name in Mode.concrete:
            custom = False
            modeClass = eval(name.capitalize())
        else:
            custom = True
            modeClass = Custom
        # Create the Mode instance
        r = modeClass(klass, className, uiSearch, inPopup, tool)
        if custom:
            # The custom PX is named "name" on p_klass
            r.px = getattr(klass, name)
        r.init(req, dir)
        return r

    @staticmethod
    def getText(name, _):
        '''Gets the i18n text corresponding to mode named p_name'''
        if '_' in name: name = name.rsplit('_', 1)[0]
        return name in Mode.concrete and _('result_mode_%s' % name) \
                                     or _('custom_%s' % name)

    def __init__(self, klass, className, uiSearch, inPopup, tool):
        # The class from which we will search instances
        self.klass = klass
        self.className = className
        # The tied UI search
        self.uiSearch = uiSearch
        # Are we in a popup ?
        self.inPopup = inPopup
        # The (z)tool
        self.tool = tool
        # The ID of the tag that will be ajax-filled with search results
        self.ajaxHookId = 'queryResult'
        # Matched objects
        self.objects = None
        # A Batch instance, when only a sub-set of the result set is shown at
        # once.
        self.batch = None
        # Are we sure the result is empty ? (ie, objects could be empty but
        # matched objects could be absent due to filters)
        self.empty = True
        # URL for triggering a new search
        self.newSearchUrl = None
        # Is the search triggered from a Ref field ?
        self.fromRef = False
        # The target for "a" tags
        self.target = tool.getLinksTargetInfo(klass)
        # How to render links to result objects ?
        self.titleMode = uiSearch.getTitleMode(tool, klass, inPopup)

    def init(self, req, dir):
        '''Every concrete class may have a specific initialization part'''
        # Store criteria for custom searches
        self.criteria = req.get('criteria')

    def getAjaxData(self):
        '''Initializes an AjaxData object on the DOM node corresponding to the
           ajax hook for this search result.'''
        # For performing a complete Ajax request, "className" is not needed
        # because included in the PX name. But it is requested by sub-Ajax
        # queries at the row level.
        search = self.uiSearch
        name = search.name
        params = {'className': self.className, 'search': name,
                  'popup': self.inPopup and '1' or '0'}
        # Add initiator-specific params
        if search.initiator:
            initatorParams = search.initiator.getAjaxParams()
            if initatorParams: params.update(initatorParams)
        # Add custom search criteria
        if self.criteria:
            params['criteria'] = self.criteria
        # Concrete classes may add more parameters
        self.updateAjaxParameters(params)
        # Convert params into a JS dict
        params = sutils.getStringFrom(params)
        px = '%s:%s:pxResult' % (self.className, name)
        return "new AjaxData('%s', '%s', %s, null, '%s', 'POST')" % \
               (self.ajaxHookId, px, params, self.tool.absolute_url())

    def updateAjaxParameters(self, params):
        '''To be overridden by subclasses for adding Ajax parameters
           (see m_getAjaxData above)'''

    def getAjaxDataRow(self, zobj, **params):
        '''Initializes an AjaxData object on the DOM node corresponding to
           p_zobj = a row (or another UI element like a cell) within the
           results.'''
        return "new AjaxData('%s', 'pxViewAsResultFromAjax', %s, '%s', '%s')"% \
               (zobj.id, sutils.getStringFrom(params), self.ajaxHookId,
                zobj.absolute_url())

    def getRefUrl(self):
        '''When the search is triggered from a Ref field, this method returns
           the URL allowing to navigate back to the page where this Ref field
           lies.'''

class List(Mode):
    '''Displays search results as a table containing one row per object'''
    px = Px('''
     <table class=":ztool.getResultCss(className, layoutType)" width="100%">
      <!-- Headers, with filters and sort arrows -->
      <tr if="showHeaders">
       <th for="column in mode.columns"
           var2="field=column.field" width=":column.width" align=":column.align"
           class=":(field == '_checkboxes') and mode.cbClass or ''">
        <x if="column.header">
         <img if="field == '_checkboxes'" src=":url('checkall')"
              class="clickable" title=":_('check_uncheck')"
              onclick=":'toggleAllCbs(%s)' % q(mode.checkboxesId)"/>
         <x if="not column.special">
          <div>::ztool.truncateText(_(field.labelId))</div>
          <!-- Sort icons -->
          <x var="sortable=field.isSortable(usage='search')"
             if="sortable and (mode.batch.total &gt; 1)">
           <img if="(mode.sortKey != field.name) or (mode.sortOrder == 'desc')"
                onclick=":'askBunchSorted(%s, %s, %s)' % \
                          (q(mode.ajaxHookId), q(field.name), q('asc'))"
                src=":url('sortDown')" class="clickable"/>
           <img if="(mode.sortKey != field.name) or (mode.sortOrder == 'asc')"
                onclick=":'askBunchSorted(%s, %s, %s)' % \
                          (q(mode.ajaxHookId), q(field.name), q('desc'))"
                src=":url('sortUp')" class="clickable"/>
          </x>
          <!-- Filter widget -->
          <x if="field.filterPx and ((mode.batch.total &gt; 1) or \
                 mode.filters)">:getattr(field, field.filterPx)</x>
          <x>:tool.pxShowDetails</x>
         </x>
        </x>
       </th>
      </tr>
      <!-- Results -->
      <tr if="not mode.objects">
       <td colspan=":len(mode.columns)+1">::_('query_no_result')</td>
      </tr>
      <x for="zobj in mode.objects"
         var2="rowCss=loop.zobj.odd and 'even' or 'odd';
              @currentNumber=currentNumber + 1">:zobj.appy().pxViewAsResult</x>
     </table>
     <!-- The button for selecting objects and closing the popup -->
     <div if="inPopup and mode.cbShown" align=":dleft">
      <input type="button"
             var="label=_('object_link_many'); css=ztool.getButtonCss(label)"
             value=":label" class=":css" style=":url('linkMany', bg=True)"
             onclick=":uiSearch.initiator.jsSelectMany(\
                   q, mode.sortKey, mode.sortOrder, mode.getFiltersString())"/>
     </div>
     <!-- Custom actions -->
     <x var="actions=uiSearch.search.getActions(tool)"
        if="actions and not inPopup">:mode.pxActions</x>
     <!-- Init checkboxes if present -->
     <script if="mode.checkboxes">:'initCbs(%s)' % q(mode.checkboxesId)</script>
     <script>:'initFocus(%s)' % q(mode.ajaxHookId)</script>''')

    def init(self, req, dir):
        '''List-specific initialization'''
        Mode.init(self, req, dir)
        search = self.uiSearch
        tool = self.tool
        # [Custom searches only] 
        if search.name == 'customSearch':
            self.refObject, self.refField = tool.getRefInfo()
            # Build the URL allowing to trigger a new search
            obj = self.refObject
            part = obj and ('&ref=%s:%s' % (obj.id, self.refField)) or ''
            self.newSearchUrl = '%s/search?className=%s%s' % \
                                (tool.absolute_url(), self.className, part)
        else:
            self.refObject, self.refField = None, None
        self.fromRef = bool(self.refField)
        # Build query parameters (start number, sort and filter)
        start = int(req.get('startNumber', '0'))
        self.sortKey = req.get('sortKey', '')
        self.sortOrder = req.get('sortOrder', 'asc')
        self.filters = tool.getFilters(self.klass)
        # Execute the query
        r = tool.executeQuery(self.className, search=search.search,
          startNumber=start, remember=True, sortBy=self.sortKey,
          sortOrder=self.sortOrder, filters=self.filters,
          refObject=self.refObject, refField=self.refField)
        # Unwrap result elements
        self.objects = r.objects
        # Create a Batch instance
        self.batch = Batch(self.ajaxHookId, r.totalNumber, len(r.objects),
                           size=r.batchSize, start=start)
        # Show sub-titles ?
        self.showSubTitles = req.get('showSubTitles', 'true') == 'true'
        # Every mached object may be selected via a checkbox
        self.rootHookId = search.getRootHookId()
        self.checkboxes = search.checkboxes
        self.checkboxesId = self.rootHookId + '_objs'
        self.cbShown = search.showCheckboxes()
        self.cbClass = not self.cbShown and 'hide' or ''
        # Determine result emptiness
        self.empty = not self.objects and not self.filters
        # Compute info related to every column in the list
        self.columnLayouts = self.getColumnLayouts()
        self.columns = tool.getColumnsSpecifiers(self.className,
                         self.columnLayouts, dir, addCheckboxes=self.checkboxes)

    def updateAjaxParameters(self, params):
        '''List-specific ajax parameters'''
        params.update(
          {'startNumber': self.batch.start, 'filters': self.filters,
           'sortKey': self.sortKey, 'sortOrder': self.sortOrder,
           'checkboxes': self.checkboxes, 'checkboxesId': self.checkboxesId,
           'totalNumber': self.batch.total,
           'resultMode': self.__class__.__name__.lower()})

    def getColumnLayouts(self):
        '''Returns the column layouts'''
        r = None
        tool = self.tool
        name = self.uiSearch.name
        # Try first to retrieve this info from a potential source Ref field
        obj = self.refObject
        if obj:
            field = obj.getAppyType(self.refField)
            r = field.getAttribute(obj, 'shownInfo')
        elif ',' in name:
            objId, fieldName, dummy = name.split(',')
            # Object may be temporary
            obj = tool.getObject(objId) or tool.getObject(objId, temp=True)
            field = obj.getAppyType(fieldName)
            if field.type == 'Ref':
                r = field.getAttribute(obj, 'shownInfo')
        if r: return r
        # Try to retrieve this info via search.shownInfo
        r = self.uiSearch.search.shownInfo
        if r: return r
        # Try now to retrieve this info via the class' "listColumns"
        k = tool.getAppyClass(self.className)
        if not hasattr(k, 'listColumns'): return ('title',)
        if callable(k.listColumns): return k.listColumns(tool.appy())
        return k.listColumns

    def getFiltersString(self):
        '''Converts dict self.filters into its string representation'''
        filters = self.filters
        if not filters: return ''
        r = []
        for k, v in filters.iteritems():
            r.append('%s:%s' % (k, v))
        return ','.join(r)

    def inFilter(self, name, value):
        '''Returns True if this p_value, for field named p_name, is among the
           currently set p_self.filters.'''
        values = self.filters.get(name)
        if not values: return
        return value in values

    def getNavInfo(self, number):
        '''Gets the navigation string corresponding to the element having this
           p_number within the list of results.'''
        return 'search.%s.%s.%d.%d' % (self.className, self.uiSearch.name, \
                                    self.batch.start + number, self.batch.total)

    def getRefUrl(self):
        '''See Mode::getRefUrl's docstring'''
        o = self.refObject.appy()
        return '%s?page=%s' % (o.url, o.getField(self.refField).page.name)

class Grid(List):
    '''Displays search results as a table containing one cell per object'''
    px = Px('''
     <div style=":mode.gridder.getContainerStyle()">
      <div for="zobj in mode.objects" class="thumbnail"
           var2="obj=zobj.appy(); mayView=zobj.mayView()">
       <table class="thumbtable">
        <tr var="@currentNumber=currentNumber + 1" valign="top"
            for="column in mode.columns"
            var2="field=column.field; backHook='queryResult'">
         <td if="field.name=='title'" colspan="2">:field.pxRenderAsResult</td>
         <x if="field.name!='title'">
          <td><label lfor=":field.name">::_('label', field=field)</label></td>
          <td>:field.pxRenderAsResult</td>
         </x>
        </tr>
       </table>
       <p class="thumbmore"><img src=":url('more')" class="clickable"
          onclick="followTitleLink(this)"/></p>
      </div>
     </div>''',

     js='''
      followTitleLink = function(img) {
        var parent = img.parentNode.parentNode,
            atag = parent.querySelector("a[name=title]");
        atag.click();
      }
      ''')

    def __init__(self, *args):
        List.__init__(self, *args)
        # Extract the gridder defined on p_self.klass or create a default one
        self.gridder = getattr(self.klass, 'gridder', None) or Gridder()

class Calendar(Mode):
    '''Displays search results in a monthly calendar view'''
    px = Px('''<x var="layoutType='view';
                       field=tool.getField('calendar')">:field.pxView</x>''')

    def __init__(self, *args):
        Mode.__init__(self, *args)
        # For this calendar view to work properly, objects to show inside it
        # must have the following attributes:
        # ----------------------------------------------------------------------
        # date    | an indexed and required Date field with format =
        #         | Date.WITH_HOUR storing the object's start date and hour;
        # ----------------------------------------------------------------------
        # endDate | a not necessarily required Date field with format =
        #         | Date.WITH_HOUR storing the object's end date and hour.
        # ----------------------------------------------------------------------
        # Optionally, if objects define the following attributes, special icons
        # indicating repeated events will be shown.
        # ----------------------------------------------------------------------
        # successor   | Ref field with multiplicity = (0,1), allowing to
        #             | navigate to the next object in the list (for a repeated
        #             | event);
        # ----------------------------------------------------------------------
        # predecessor | Ref field with multiplicity = (0,1), allowing to
        #             | navigate to the previous object in the list.
        #             | "predecessor" must be the back ref of field "successor"
        #             | hereabove.
        # ----------------------------------------------------------------------

    def init(self, req, dir):
        '''Creates a stub calendar field'''
        Mode.init(self, req, dir)
        # Always consider the result as not empty. This way, the calendar is
        # always shown, even if no object is visible.
        self.empty = False
        # The matched objects, keyed by day. For every day, a list of entries to
        # show. Every entry is a 2-tuple (s_entryType, Object) allowing to
        # display an object at this day. s_entryType can be:
        # ----------------------------------------------------------------------
        #  "start"  | the object starts and ends at this day: display its start
        #           | hour and title;
        # ----------------------------------------------------------------------
        #  "start+" | the object starts at this day but ends at another day:
        #           | display its start hour, title and some sign indicating
        #           | that it spans on another day;
        # ----------------------------------------------------------------------
        #  "end"    | the object started at a previous day and ends at this day.
        #           | Display its title and a sign indicating this fact;
        # ----------------------------------------------------------------------
        #  "pass"   | The object started at a previous day and ends at a future
        #           | day.
        # ----------------------------------------------------------------------
        self.objects = {} # ~{s_YYYYmmdd: [(s_entryType, Object)]}~
        # Formats for keys representing dates
        self.dayKey = '%Y%m%d'
        # Format for representing hours in the UI
        self.hourFormat = '%H:%M'
        # If filters are defined from a list mode, get it
        self.filters = self.tool.getFilters(self.klass)

    def updateAjaxParameters(self, params):
        '''Grid-specific ajax parameters'''
        # If filters are in use, carry them
        if self.filters:
            params['filters'] = self.filters

    # For every hereabove-defined entry type, this dict stores info about how
    # to render events having this type. For every type:
    # --------------------------------------------------------------------------
    # start      | bool | Must we show the event's start hour or not ?
    # end        | bool | Must we show the event's end hour or not ?
    # css        | str  | The CSS class to add the table event tag
    # past       | bool | True if the event spanned more days in the past
    # future     | bool | True if the event spans more days in the future
    # --------------------------------------------------------------------------
    entryTypes = {
     'start':  O(start=True,  end=True,  css=None,      past=None, future=None),
     'start+': O(start=True,  end=False, css='calMany', past=None, future=True),
     'end':    O(start=False, end=True,  css='calMany', past=True, future=None),
     'pass':   O(start=False, end=False, css='calMany', past=True, future=True),
    }

    def addEntry(self, dateKey, entry):
        '''Adds an p_entry as created by m_addEntries below into self.objects
           @key p_dateKey.'''
        r = self.objects
        if dateKey not in r:
            r[dateKey] = [entry]
        else:
            r[dateKey].append(entry)

    def addEntries(self, obj):
        '''Add, in self.objects, entries corresponding to p_obj. If p_obj spans
           a single day, a single entry of the form ("start", p_obj) is added at
           the key corresponding to this day. Else, a series of entries are
           added, each of the form (s_entryType, p_obj), with the same object,
           for every day in p_obj's timespan.

           For example, for an p_obj starting at "1975/12/11 12:00" and ending
           at "1975/12/13 14:00" will produce the following entries:
              key "19751211"  >  value ("start+", obj)
              key "19751212"  >  value ("pass", obj)
              key "19751213"  >  value ("end", obj)
        '''
        # Get p_obj's start and end dates
        start = obj.date
        startKey = start.strftime(self.dayKey)
        end = obj.endDate
        endKey = end and end.strftime(self.dayKey) or None
        # Shorthand for self.objects
        r = self.objects
        if not endKey or (endKey == startKey):
            # A single entry must be added for p_obj, at the start date
            self.addEntry(startKey, ("start", obj))
        else:
            # Add one entry at the start day
            self.addEntry(startKey, ("start+", obj))
            # Add "pass" entries for every day between the start and end days
            next = start + 1
            nextKey = next.strftime(self.dayKey)
            while nextKey != endKey:
                # Add a "pass" event
                self.addEntry(nextKey, ('pass', obj))
                # Go the the next day
                next += 1
                nextKey = next.strftime(self.dayKey)
            # Add an "end" entry at the end day
            self.addEntry(endKey, ('end', obj))

    def search(self, first, grid):
        '''Performs the query, limited to the date range defined by p_grid'''
        # Performs the query, restricted to the visible date range
        last = DateTime(grid[-1][-1].strftime('%Y/%m/%d 23:59:59'))
        dateSearch = Search(date=(first, last), sortBy='date', sortOrder='asc')
        res = self.tool.executeQuery(self.className,
          search=self.uiSearch.search, maxResults='NO_LIMIT',
          search2=dateSearch, filters=self.filters)
        # Produce, in self.objects, the dict of matched objects
        for zobj in res.objects:
            self.addEntries(zobj.appy())

    def dumpObjectsAt(self, date):
        '''Returns info about the object(s) that must be shown in the cell
           corresponding to p_date.'''
        # There may be no object dump at this date
        dateStr = date.strftime(self.dayKey)
        if dateStr not in self.objects: return
        # Objects exist
        r = []
        types = self.entryTypes
        url = self.tool.getIncludeUrl
        for entryType, obj in self.objects[dateStr]:
            # Dump the event hour and title. The back hook allows to refresh the
            # whole calendar view when coming back from the popup.
            eType = types[entryType]
            # What CSS class(es) to apply ?
            css = eType.css and ('calEvt %s' % eType.css) or 'calEvt'
            # Show start and/or end hour ?
            eHour = sHour = ''
            if eType.start:
                sHour = '<td width="2em">%s</td>' % \
                        obj.date.strftime(self.hourFormat)
            if eType.end:
                endDate = obj.endDate
                if endDate:
                    eHour = ' <abbr title="%s">¬</abbr>' % \
                            endDate.strftime(self.hourFormat)
            # Display indicators that the event spans more days
            past = eType.past and '⇠ ' or ''
            future = eType.future and ' ⇢' or ''
            # The event title
            title = obj.o.getListTitle(target=self.target, inPopup=True,
              backHook='configcalendar', maxChars=24)
            # Display a "repetition" icon if the object is part of a series
            hasSuccessor = obj.ids('successor')
            hasPredecessor = obj.ids('predecessor')
            if hasSuccessor or hasPredecessor:
                # For the last event of a series, show a more stressful icon
                name = not hasSuccessor and 'repeated_last' or 'repeated'
                icon = '<img src="%s" class="help" title="%s"/>' % \
                       (url(name), obj.translate(name))
            else:
                icon = ''
            # Produce the complete entry
            r.append('<table class="%s"><tr valign="top">%s<td>%s%s%s%s%s</td>'\
                     '</tr></table>' % (css,sHour,past,title,future,eHour,icon))
        return '\n'.join(r)

class Custom(Mode):
    '''Displays search results via a custom PX'''
    def init(self, req, dir):
        '''By default, the Custom mode performs full (unpaginated) searches'''
        Mode.init(self, req, dir)
        r = self.tool.executeQuery(self.className, search=self.uiSearch.search,
                                   maxResults='NO_LIMIT')
        # Initialise Mode's mandatory fields
        self.objects = r.objects
        self.empty = not self.objects

# ------------------------------------------------------------------------------
class ColSet:
    '''Represents a named set of columns to show when displaying Search results
       (or also Refs).'''
    def __init__(self, identifier, label, columns, specs=False):
        # A short identifier for the set
        self.identifier = identifier
        # The i18n label to use for giving a human-readable name to the set
        self.label = label
        # The list/tuple of columns, expressed as strings. Every string must
        # contain a field name, but can be completed (after a char *) by column
        # width and alignment, as in "state*100px|". The "width" part, just
        # after the *, can hold anything that can be set in a "width" HTML
        # attribute. The last char represents the alignment:
        #   ";"   left-aligned (the default);
        #   "|"   centered;
        #   "!"   right-aligned.
        if not specs:
            self.columns = columns
        else:
            # "specs" is the internal representation of "columns". Do not
            # specify "specs=True". It will contain a list of Object instances
            # instead of strings. Every such instance has splitted string info
            # into fields "field", "width" and "align".
            self.specs = columns

# ------------------------------------------------------------------------------
class SearchInitiator:
    '''Initiator in use when an object is created from a search'''

    # A Search initiator cannot be used as folder in the context of object
    # creation.
    asFolder = False

    def __init__(self, tool, req, info):
        self.tool = tool
        self.req = req
        # Extract the initiator object and field from p_info, parsed from a
        # "nav" key in the request.
        self.info = info.split('.')
        # Get info about the field from which the object creation was triggered
        if ',' in self.info[1]:
            objectId, fieldName, layout = self.info[1].split(',')
            self.obj = tool.getObject(objectId)
            self.field = self.obj.getField(fieldName)
        else:
            # There is no initiator
            self.obj = self.field = None
        self.backFromPopupUrl = None

    # Conform to the Initiator API
    def checkAllowed(self): pass
    def updateParameters(self, urlParams): pass
    def manage(self, new): pass
    def goBack(self): return 'view'

    def getUrl(self):
        '''Returns the URL for going back to the initiator object, on the page
           showing self.field.'''
        return self.obj.o.getUrl(page=self.field.pageName, nav='no')

    def getNavInfo(self, new):
        '''Reconstitute the navInfo for navigating from p_new'''
        # It is impossible to know at what position p_new will come within the
        # originating search. By specifying the total number of elements being
        # 0, it deactivates sibling navigation and keeps only the icon to go
        # back to the complete search.
        return 'search.%s.%s.0.0' % (self.info[0], self.info[1])

    def isComplete(self):
        '''The initiator is considered to be complete if "obj" is not empty'''
        return self.obj

# ------------------------------------------------------------------------------
class Search:
    '''Used for specifying a search for a given class'''

    # Make this class available as "Search.ColSet"
    ColSet = ColSet

    # Special initiator in use when an object is being created from a Search. It
    # has nothing to do with popup-oriented initiator as defined hereafter.
    initiator = SearchInitiator

    def __init__(self, name=None, group=None, sortBy='', sortOrder='asc',
                 maxPerPage=30, default=False, colspan=1, translated=None,
                 show=True, showActions='all', actionsDisplay='block',
                 translatedDescr=None, checkboxes=False, checkboxesDefault=True,
                 klass=None, add=False, resultModes=None, shownInfo=None,
                 actions=None, **fields):
        # "name" is mandatory, excepted in some special cases (ie, when used as
        # "select" param for a Ref field).
        self.name = name
        # Searches may be visually grouped in the portlet
        self.group = Group.get(group)
        self.sortBy = sortBy
        self.sortOrder = sortOrder
        self.maxPerPage = maxPerPage
        # If this search is the default one, it will be triggered by clicking
        # on main link.
        self.default = default
        self.colspan = colspan
        # If a translated name or description is already given here, we will
        # use it instead of trying to translate from labels.
        self.translated = translated
        self.translatedDescr = translatedDescr
        # Condition for showing or not this search
        self.show = show
        # Attributes "showActions" and "actionsDisplay" are similar to their
        # homonyms on the Ref class.
        self.showActions = showActions
        self.actionsDisplay = actionsDisplay
        # In the dict below, keys are indexed field names or names of standard
        # indexes, and values are search values.
        self.fields = fields
        # Do we need to display checkboxes for every object of the query result?
        self.checkboxes = checkboxes
        # Default value for checkboxes
        self.checkboxesDefault = checkboxesDefault
        # Most of the time, we know what is the class whose instances must be
        # searched. When it is not the case, the p_klass can be explicitly
        # specified.
        self.klass = klass
        # Is it possible to create new "klass" instances from this search ?
        self.add = add
        # There can be various ways to display query results
        self.resultModes = resultModes
        # Similar to the homonym Ref attribute, "shownInfo" defines the columns
        # that must be shown on lists of result objects (mode "List" only). If
        # not specified, class's "listColumns" attributes is used.
        self.shownInfo = shownInfo
        # Specify here Action fields that must be shown as custom actions that
        # will be triggered on search results.
        self.actions = actions

    @staticmethod
    def getIndexName(name, klass, usage='search'):
        '''Gets the name of the Zope index that corresponds to p_name. Indexes
           can be used for searching (p_usage="search"), for filtering
           ("filter") or for sorting ("sort"). The method returns None if the
           field named p_name can't be used for p_usage.'''
        # Manage indexes that do not have a corresponding field
        if name == 'created': return 'Created'
        elif name == 'modified': return 'Modified'
        elif name in defaultIndexes: return name
        else:
            # Manage indexes corresponding to fields
            field = getattr(klass, name, None) 
            if field: return field.getIndexName(usage)
            raise Exception(WRONG_FIELD % (name, klass.__bases__[-1].__name__))

    @staticmethod
    def getSearchValue(fieldName, fieldValue, klass):
        '''Returns a transformed p_fieldValue for producing a valid search
           value as required for searching in the index corresponding to
           p_fieldName.'''
        # Get the field corresponding to p_fieldName
        field = getattr(klass, fieldName, None)
        if field and callable(field): field = None
        if (field and (field.getIndexType() == 'TextIndex')) or \
           (fieldName == 'SearchableText'):
            # For TextIndex indexes. We must split p_fieldValue into keywords.
            res = gutils.Keywords(fieldValue).get()
        elif isinstance(fieldValue, basestring) and fieldValue.endswith('*'):
            v = fieldValue[:-1]
            # Warning: 'z' is higher than 'Z'!
            res = {'query':(v,v+'z'), 'range':'min:max'}
        elif type(fieldValue) in sutils.sequenceTypes:
            if fieldValue and isinstance(fieldValue[0], basestring):
                # We have a list of string values (ie: we need to
                # search v1 or v2 or...)
                res = fieldValue
            else:
                # We have a range of (int, float, DateTime...) values
                minv, maxv = fieldValue
                rangev = 'minmax'
                queryv = fieldValue
                if minv == None:
                    rangev = 'max'
                    queryv = maxv
                elif maxv == None:
                    rangev = 'min'
                    queryv = minv
                res = {'query':queryv, 'range':rangev}
        else:
            res = fieldValue
        return res

    def updateSearchCriteria(self, tool, criteria, klass, advanced=False):
        '''This method updates dict p_criteria with all the search criteria
           corresponding to this Search instance. If p_advanced is True,
           p_criteria correspond to an advanced search, to be stored in the
           session: in this case we need to keep the Appy names for parameters
           sortBy and sortOrder (and not "resolve" them to Zope's sort_on and
           sort_order).'''
        # Beyond parameters required for performing a search, also store, in
        # p_criteria, other Search parameters if we need to reify a Search
        # instance for performing an advanced search.
        if advanced:
            cb = self.checkboxes
            if callable(cb): cb = cb(tool.appy())
            criteria['showActions'] = self.showActions
            criteria['actionsDisplay'] = self.actionsDisplay
            criteria['resultModes'] = self.resultModes
            criteria['shownInfo'] = self.shownInfo
            criteria['checkboxes'] = cb
        # Put search criteria in p_criteria
        for name, value in self.fields.iteritems():
            # Management of searches restricted to objects linked through a
            # Ref field: not implemented yet.
            if name == '_ref': continue
            # Make the correspondence between the name of the field and the
            # name of the corresponding index, excepted if p_advanced is True:
            # in that case, the correspondence will be done later.
            if not advanced:
                indexName = Search.getIndexName(name, klass)
                # Express the field value in the way needed by the index
                criteria[indexName] = Search.getSearchValue(name, value, klass)
            else:
                criteria[name] = value
        # Add a sort order if specified
        if self.sortBy:
            c = criteria
            if not advanced:
                c['sort_on']=Search.getIndexName(self.sortBy,klass,usage='sort')
                c['sort_order']= (self.sortOrder=='desc') and 'reverse' or None
            else:
                c['sortBy'] = self.sortBy
                c['sortOrder'] = self.sortOrder

    def isShowable(self, klass, tool):
        '''Is this Search instance (defined in p_klass) showable ?'''
        if self.show.__class__.__name__ == 'staticmethod':
            return gutils.callMethod(tool, self.show, klass=klass)
        return self.show

    def getSessionKey(self, className, full=True):
        '''Returns the name of the key, in the session, where results for this
           search are stored when relevant. If p_full is False, only the suffix
           of the session key is returned (ie, without the leading
           "search_").'''
        res = (self.name == 'allSearch') and className or self.name
        if not full: return res
        return 'search_%s' % res

    mergeFields = ('sortBy', 'sortOrder', 'showActions',
                   'actionsDisplay', 'actions')

    def merge(self, other):
        '''Merge parameters from another search in p_other'''
        self.fields.update(other.fields)
        for name in self.mergeFields: setattr(self, name, getattr(other, name))

    def run(self, tool):
        '''Executes this query. Works only when self.klass exists.'''
        if not self.klass:
            raise Exception('Running this query requires self.klass')
        return tool.search(self.klass, sortBy=self.sortBy,
                           sortOrder=self.sortOrder, **self.fields)

    def getActions(self, tool):
        '''Get the actions triggerable on p_self's results'''
        actions = self.actions
        if not actions: return
        r = []
        for action in actions:
            show = action.show
            if callable(show): show = show(tool)
            if show:
                r.append(action)
        return r

# Initiators for searches whose results are shown in popups --------------------
class Initiator:
    '''When a query is rendered in a popup, the "initiator", in the main page,
       can be:
       * (a) some object, in view or edit mode, displaying a given Ref field
             for which the popup is used to select one or more objects to link;
       * (b) some class for which we must create an instance from a template;
             the popup is used to select such a template object.

       This class is the abstract class for 2 concrete initiator classes:
       RefInitiator (for case a) and TemplateInitiator (for case b).
    '''

class RefInitiator(Initiator):
    def __init__(self, obj, field, fieldName, mode):
        # The initiator object
        self.obj = obj
        # The initiator field
        self.field = field
        # As usual, the field name can be different from field.name if it is a
        # sub-field within a List field
        self.fieldName = fieldName
        # The mode can be:
        # - "repl" if the objects selected in the popup will replace already
        #          tied objects;
        # - "add"  if those objects will be added to the already tied ones.
        self.mode = mode
        # "hook" is the ID of the initiator field's XHTML tag
        self.hook = '%s_%s' % (obj.id, fieldName)
        # The root Ajax hook ID in the popup
        self.popupHook = '%s_popup' % self.hook

    def showCheckboxes(self):
        '''We must show object checkboxes if self.field is multivalued: indeed,
           in this case, several objects can be selected in the popup.'''
        return self.field.isMultiValued()

    def jsSelectOne(self, q, cbId):
        '''Generates the Javascript code to execute when a single object is
           selected in the popup.'''
        return 'onSelectObject(%s,%s,%s)' % \
               (q(cbId), q(self.hook), q(self.obj.url))

    def jsSelectMany(self, q, sortKey, sortOrder, filters):
        '''Generates the Javascript code to execute when several objects are
           selected in the popup.'''
        return 'onSelectObjects(%s,%s,%s,%s,null,%s,%s,%s)' % \
          (q(self.popupHook), q(self.hook), q(self.obj.url), q(self.mode), \
           q(sortKey), q(sortOrder), q(filters))

    def getAjaxParams(self):
        '''Get initiator-specific parameters for retriggering the Ajax
           request for refreshing objects in the popup.'''
        return

class TemplateInitiator(Initiator):
    MANY_ERROR = 'Cannot select several objects from a template initiator.'

    def __init__(self, className, formName, insert, sourceField):
        # The class from which we must create an instance based on a template
        # that we will choose in the popup. Indeed, the instance to create may
        # be from a different class that the instances shown in the popup.
        self.className = className
        # The name of the form that will be submitted for creating the object
        # once a template will have been selected in the popup.
        self.formName = formName
        # The root Ajax hook ID in the popup
        self.popupHook = '%s_popup' % className
        # If the object to create must be inserted at a given place in a Ref
        # field, this can be specified in p_insert.
        self.insert = insert or ''
        # The source field
        self.sourceField = sourceField

    def showCheckboxes(self):
        '''We must hide object checkboxes: only one template object can be
           selected.'''
        return

    def jsSelectOne(self, q, cbId):
        '''Generates the Javascript code to execute when a single object is
           selected in the popup.'''
        return 'onSelectTemplateObject(%s,%s,%s)' % \
               (q(cbId), q(self.formName), q(self.insert))

    def jsSelectMany(self, q, sortKey, sortOrder, filters):
        raise Exception(self.MANY_ERROR)

    def getAjaxParams(self):
        res = {'fromClass': self.className, 'formName': self.formName}
        if self.insert:
            res['insert'] = self.insert
        if self.sourceField:
            res['sourceField'] = self.sourceField
        return res

# ------------------------------------------------------------------------------
class UiSearch:
    '''Instances of this class are generated on-the-fly for manipulating a
       Search instance from the User Interface.'''
    # Tied sub-classes
    Mode = Mode
    Calendar = Calendar
    RefInitiator = RefInitiator
    TemplateInitiator = TemplateInitiator

    # Rendering a search
    pxView = Px('''
     <div class="portletSearch">
      <a href=":'%s?className=%s&amp;search=%s' % \
                 (queryUrl, className, search.name)"
         class=":(search.name == currentSearch) and 'current' or ''"
         onclick="clickOn(this)"
         title=":search.translatedDescr">:search.translated</a>
     </div>''')

    # Render search results
    pxResult = Px('''
     <x var="layoutType='view';
             className=req['className'];
             klass=ztool.getAppyClass(className);
             uiSearch=uiSearch|ztool.getSearch(className,req['search'],ui=True);
             mode=uiSearch.Mode.get(klass, className, uiSearch, inPopup, \
                                    ztool, req, dir);
             batch=mode.batch;
             empty=mode.empty;
             showNewSearch=showNewSearch|True;
             showHeaders=showHeaders|True;
             specific=uiSearch.getResultsTop(tool, klass, mode, ajax)">

     <!-- Application, class-specific code before displaying results -->
     <x if="specific">::specific</x>

     <!-- Search results -->
     <div id=":mode.ajaxHookId">
      <script>:mode.getAjaxData()</script>

      <!-- Pod templates -->
      <x if="not empty and not inPopup">
       <table var="fields=ztool.getResultPodFields(className)"
              if="fields and mode.objects" align=":dright">
        <tr>
         <td var="zobj=mode.objects[0]; obj=zobj.appy()"
             for="field in fields" var2="fieldName=field.name"
             class=":not loop.field.last and 'pod' or ''">:field.pxRender</td>
        </tr>
       </table>
      </x>

      <!-- Title -->
      <div if="not inPopup" class="refBar"><x>::uiSearch.translated</x>
       <x if="mode.batch and not empty">(
        <span class="discreet">:mode.batch.total</span>)
       </x>
       <x if="uiSearch.search.add and ztool.userMayCreate(klass)"
          var2="rootClass=klass; asButton=True;
                nav=mode.getNavInfo(batch.total)">:tool.pxAdd</x>
       <x if="showNewSearch and mode.newSearchUrl">&nbsp;&mdash;&nbsp;<i> 
        <a href=":mode.newSearchUrl">:_('search_new')</a></i></x>
       <x if="mode.fromRef">&nbsp;&mdash;&nbsp;<i> 
        <a href=":mode.getRefUrl()">:_('goto_source')</a></i></x>
      </div>
      <table width="100%">
       <tr valign="top">
        <!-- Search description -->
        <td if="uiSearch.translatedDescr">
         <span class="discreet">:uiSearch.translatedDescr</span><br/>
        </td>
        <!-- (Top) navigation -->
        <td if="mode.batch" align=":dright">:mode.batch.pxNavigate</td>
       </tr>
      </table>

      <!-- Results -->
      <x if="not empty" var2="currentNumber=0">:mode.px</x>

      <!-- (Bottom) navigation -->
      <x if="mode.batch">:mode.batch.pxNavigate</x>

      <!-- No result -->
      <x if="empty">
       <x>::_('query_no_result')</x>
       <x if="showNewSearch and mode.newSearchUrl"><br/><i class="discreet">
         <a href=":mode.newSearchUrl">:_('search_new')</a></i></x>
      </x>
    </div></x>''')

    def __init__(self, search, className, tool, initiator=None, name=None):
        self.search = search
        # "name" can be more than the p_search name, ie, if the search is
        # defined in a field.
        self.name = name or search.name
        self.type = 'search'
        self.colspan = search.colspan
        self.className = className
        self.showActions = search.showActions
        self.actionsDisplay = search.actionsDisplay
        if search.translated:
            self.translated = search.translated
            self.translatedDescr = search.translatedDescr or ''
        else:
            # The label may be specific in some special cases
            labelDescr = ''
            if search.name == 'allSearch': label = '%s_plural' % className
            elif search.name == 'customSearch': label = 'search_results'
            elif (not search.name or search.klass): label = None
            else:
                label = '%s_search_%s' % (className, search.name)
                labelDescr = label + '_descr'
            _ = tool.translate
            self.translated = label and _(label) or ''
            self.translatedDescr = labelDescr and _(labelDescr) or ''
        # Strip the description (a single space may be present)
        self.translatedDescr = self.translatedDescr.strip()
        # An initiator instance if the query is in a popup
        self.initiator = initiator
        # When query results are shown in a popup, checkboxes must be present
        # even when not shown. Indeed, we want them in the DOM because object
        # ids are stored on it.
        if initiator:
            self.checkboxes = True
            self.checkboxesDefault = False
        else:
            cb = search.checkboxes
            if callable(cb): cb = cb(tool.appy())
            self.checkboxes = cb
            self.checkboxesDefault = search.checkboxesDefault

    def getRootHookId(self):
        '''If there is an initiator, return the hook as defined by it. Else,
           return the name of the search.'''
        if not self.initiator:
            return self.search.name or 'search'
        else:
            return self.initiator.popupHook

    def showCheckboxes(self):
        '''When must checkboxes be shown ?'''
        if not self.initiator: return self.checkboxes
        return self.initiator.showCheckboxes()

    def getCbJsInit(self, hookId):
        '''Returns the code that creates JS data structures for storing the
           status of checkboxes for every result of this search.'''
        default = self.checkboxesDefault and 'unchecked' or 'checked'
        return '''var node=findNode(this, '%s');
                  node['_appy_objs_cbs'] = {};
                  node['_appy_objs_sem'] = '%s';''' % (hookId, default)

    def getModes(self, klass, tool):
        '''Gets all the modes applicable when displaying query results (being
           instances of p_klass) via this search (p_self). r_ is a list of names
           and not a list of Mode instances.'''
        r = self.search.resultModes or \
            getattr(klass, 'resultModes', Mode.default)
        if callable(r): r = r(tool.appy())
        return r

    def getTitleMode(self, tool, klass, inPopup):
        '''How titles to search results, being instances of p_klass, must be
           rendered ? r_ is:
           * "link" : as links allowing to go to instances' view pages;
           * "select": as objects that can be selected from a popup;
           * "text": as simple, unclickable text.
        '''
        if inPopup: return 'select'
        # Check if the title mode is specified on p_klass
        mode = getattr(klass, 'titleMode', None)
        if not mode: return 'link'
        # If it is a 
        if callable(mode): mode = mode(tool.appy())
        return mode

    def getResultsTop(self, tool, klass, mode, ajax):
        '''If p_klass defines something to display on the results page just
           before displaying search results, returns it.'''
        # Get this only on the main page, not when ajax-refreshing search
        # results.
        if ajax: return
        if hasattr(klass, 'getResultsTop'):
            return klass.getResultsTop(tool, self.search, mode)

# ------------------------------------------------------------------------------
class Criteria:
    '''Represents a set of search criteria from a custom search'''

    def __init__(self, tool):
        self.tool = tool
        # This attribute will store the dict of search criteria, ready to be
        # injected in a Search class for performing a search in the catalog.
        self.criteria = None

    @staticmethod
    def readFromRequest(tool):
        '''Unmarshalls, from request key "criteria", a dict that was marshalled
           from a dict similar to the one stored in attribute "criteria" in
           Criteria instances.'''
        # Criteria may be cached in the request
        req = tool.REQUEST
        if hasattr(req, 'searchCriteria'): return req.searchCriteria
        # Criteria may be absent
        criteria = req.get('criteria')
        if not criteria: return
        # Criteria are present by not cached. Get them from the request,
        # unmarshal and cache them.
        r = eval(criteria)
        req.searchCriteria = r
        return r

    def getDefault(self, form):
        '''Get the default search criteria that may be defined on the
           corresponding Appy class, in field Class.searchAdvanced (field
           values, sort filters, etc), and return it if found.'''
        r = {}
        if 'className' not in form: return r
        # Get the Appy class for which a search is requested
        className = form['className']
        tool = self.tool
        klass = tool.getAppyClass(className)
        # On this Appy class, retrieve the Search instance containing default
        # search criteria.
        search = tool.getSearchAdvanced(klass)
        if not search: return r
        wrapperClass = tool.getAppyClass(className, wrapper=True)
        search.updateSearchCriteria(tool, r, wrapperClass, advanced=True)
        return r

    def getFromRequest(self):
        '''Retrieve search criteria from the request after the user has filled
           an advanced search form and perform some transforms on it to produce
           p_self.criteria.'''
        req = self.tool.REQUEST
        form = req.form
        # Start by collecting default search criteria
        r = self.getDefault(form)
        className = form['className']
        # Then, retrieve criteria from the request
        for name in form.keys():
            # On the Appy advanced search form, every search field is prefixed
            # with "w_".
            if not name.startswith('w_'): continue
            name = name[2:]
            # Get the field corresponding to request key "name"
            field = self.tool.getAppyType(name, className)
            # Ignore this value if it is empty or if the field is inappropriate
            # for a search.
            if not field or field.searchValueIsEmpty(form) or not field.indexed:
                continue
            # We have a(n interval of) value(s) that is not empty for a given
            # field. Get it.
            r[name] = field.getSearchValue(form)
        # Complete criteria with Ref info if the search is restricted to
        # referenced objects of a Ref field.
        refInfo = req.get('ref', None)
        if refInfo: r['_ref'] = refInfo
        self.criteria = r

    def asString(self):
        '''Returns p_self.criteria, marshalled in a string'''
        return sutils.getStringFrom(self.criteria, stringify=False)
# ------------------------------------------------------------------------------



'''Module implementing Javascript-related functionality'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import re

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Quote:
    '''This class escapes strings to be integrated as Javascript string literals
       in Javascript code, and, once escaped, quotes them.'''
    # The "escaping" part of this class is inspired by class appy/xml/Escape
    rex = re.compile("[\n\t\r']")
    blanks = {'\n': '', '\t':'', '\r':''}
    # There are 2 ways to escape a single quote
    values = {True: {"'": '&apos;'}, False: {"'": "\\'"}}
    for d in values.values(): d.update(blanks)

    @staticmethod
    def js(s, withEntity=True):
        '''Escapes blanks and single quotes in string p_s. Single quotes are
           escaped with a HTML entity if p_withEntity is True or with a quote
           prefixed with a "backslash" char else. Returns p_s, escaped and
           quoted.'''
        s = s if isinstance(s, str) else str(s)
        fun = lambda match: Quote.values[withEntity][match.group(0)]
        return "'%s'" % Quote.rex.sub(fun, s)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import re

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Escape:
    '''Escapes XML chars within strings'''

    # Escaping consists in replacing special chars used by the XML language by
    # replacement entities that can be used in the content of XML tags and
    # attributes. This class does it for various XML flavours:
    # --------------------------------------------------------------------------
    #   "odf"    | Open Document Format - The XML format for LibreOffice files
    #  "xhtml"   | The XML-compliant version of HTML
    #   "xml"    | Any other XML flavour
    # --------------------------------------------------------------------------
    # Base chars to escape
    chars = '<>&"'
    # For some XML flavours we escape "blanks" as well
    blanks = '\n\t\r'
    # Base regular expression matching these chars
    rex = re.compile('[%s]' % chars)
    # Regular expression also matching the single quote (less used)
    rexApos = re.compile("[%s']" % chars)
    # Regular expression also matching carriage returns and tabs (= "blanks")
    rexBlanks = re.compile('[%s%s]' % (chars, blanks))
    rexBlanksApos = re.compile("[%s'%s]" % (chars, blanks))
    # All regular expressions, grouped by flavour and 'apos' escaping or not
    nonXmlRex = { False: rexBlanks, True: rexBlanksApos }
    rexAll = {'xml'   : { False: rex, True: rexApos },
              'odf'   : nonXmlRex,
              'odf*'  : nonXmlRex,
              'xhtml' : nonXmlRex,
              'xhtml*': nonXmlRex}
    # Entities to use to escape base chars
    entities = {'<':'&lt;', '>':'&gt;', '&':'&amp;', '"':'&quot;', "'":'&apos;'}
    # For "odf" and "xhtml" flavours, we replace "blank" chars with their
    # counterparts in these flavours as well.
    values = {
     # While using an additional "text:tab" when replacing "\n" chars for
     # conversion to ODF solves problems with justified text, it causes problems
     # when converting the result to Microsoft Word. This is why there are 2
     # possibilities for ODF.
     'odf': {'\n':'<text:line-break/>', '\t':'<text:tab/>', '\r':''},
     'odf*':{'\n':'<text:tab/><text:line-break/>', '\t':'<text:tab/>', '\r':''},
     'xhtml': {'\n':'<br/>', '\t':'', '\r':''},
     'xhtml*': {'\n':'</p><p>', '\t':'', '\r':''},
     'xml': {} # We do not escape blanks by default
    }
    # Complete "values" with entities
    for value in values.values(): value.update(entities)

    @staticmethod
    def xml(s, flavour='xml', escapeApos=False):
        '''Returns p_s, whose XML special chars have been replaced with XML
           entities. You can perform escaping for a specific XML flavour: "odf"
           or "xhtml" (see doc in class Escape) or use the default flavour
           "xml".

           Most of the time, we do not escape 'apos' (there is no particular
           need for that), excepted if p_escapeApos is True.'''
        # Choose the regular expression to use depending on parameters
        rex = Escape.rexAll[flavour][escapeApos]
        # Define the function to use to choose the replacement value depending
        # on the matched char.
        fun = lambda match: Escape.values[flavour][match.group(0)]
        # Apply it
        return rex.sub(fun, s)

    @staticmethod
    def xhtml(s, p=False):
        '''Shorthand for escaping XHTML content'''
        # If p_p is False, \n are replaced with <br/> tags. Else, every
        # encountered \n leads to the creation of a new paragraph, surrounded by
        # <p></p>.
        suffix = '*' if p else ''
        r = Escape.xml(s, flavour='xhtml%s' % suffix)
        return '<p>%s</p>' % r if p else r
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Imports ----------------------------------------------------------------------
import os, os.path
from appy.shared import appyPath
from appy.shared.utils import FolderDeleter, cleanFolder

# ------------------------------------------------------------------------------
class Cleaner:
    def run(self, verbose=True):
        cleanFolder(appyPath, verbose=verbose)
        # Remove all files in temp folders
        for tempFolder in ('%s/temp' % appyPath,
                           '%s/pod/test/temp' % appyPath):
            if os.path.exists(tempFolder):
                FolderDeleter.delete(tempFolder)
        # Remove test reports if any
        for testReport in ('%s/pod/test/Tester.report.txt' % appyPath,):
            if os.path.exists(testReport):
                os.remove(testReport)

# Main program -----------------------------------------------------------------
if __name__ == '__main__':
    Cleaner().run()
# ------------------------------------------------------------------------------



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import types, string, copy
from group import Group
from appy.px import Px
from appy.shared.utils import sequenceTypes
from appy.gen.utils import User

# Default Appy permissions -----------------------------------------------------
r, w, d = ('read', 'write', 'delete')
emptyDict = {}
class WorkflowException(Exception): pass

# ------------------------------------------------------------------------------
class Role:
    '''Represents a role, be it local or global'''
    appyRoles = ('Manager', 'Owner', 'Anonymous', 'Authenticated')
    appyLocalRoles = ('Owner',)
    appyUngrantableRoles = ('Anonymous', 'Authenticated')
    def __init__(self, name, local=False, grantable=True):
        self.name = name
        self.local = local # True if it can be used as local role only.
        # It is a standard Zope role or an application-specific one?
        self.appy = name in self.appyRoles
        if self.appy and (name in self.appyLocalRoles):
            self.local = True
        self.grantable = grantable
        if self.appy and (name in self.appyUngrantableRoles):
            self.grantable = False
        # An ungrantable role is one that is, like the Anonymous or
        # Authenticated roles, automatically attributed to a user.

    def __repr__(self):
        loc = self.local and ' (local)' or ''
        return '<%s%s>' % (self.name, loc)

# ------------------------------------------------------------------------------
class State:
    '''Represents a workflow state'''
    def __init__(self, permissions, initial=False, phase=None, show=True):
        self.usedRoles = {}
        # Dict "permissions" lists, for every permission managed by a workflow,
        # the roles for which the permission is granted in this state. Standard
        # permissions are 'read', 'write' and 'delete'. p_permissions must be a
        # dict with format
        #            ~{s_permissionName:[s_roleName|Role_role]}~

        # It will be converted by m_standardizeRoles into format
        #            ~{s_permissionName:{s_roleName:Role_role}}~
        self.permissions = permissions
        self.initial = initial
        self.phase = phase
        self.show = show
        # Standardize the way roles are expressed within self.permissions
        self.standardizeRoles()

    def getName(self, wf):
        '''Returns the name for this state in workflow p_wf'''
        for name in dir(wf):
            value = getattr(wf, name)
            if (value == self): return name

    def copyPerms(self):
        '''Gets a deep copy of this state's permissions dict'''
        return copy.deepcopy(self.permissions)

    def getRole(self, role):
        '''p_role can be the name of a role or a Role instance. If it is the
           name of a role, this method returns self.usedRoles[role] if it
           exists, or creates a Role instance, puts it in self.usedRoles and
           returns it else. If it is a Role instance, the method stores it in
           self.usedRoles if it is not in it yet and returns it.'''
        if isinstance(role, basestring):
            if role in self.usedRoles:
                return self.usedRoles[role]
            else:
                theRole = Role(role)
                self.usedRoles[role] = theRole
                return theRole
        else:
            if role.name not in self.usedRoles:
                self.usedRoles[role.name] = role
            return role

    def standardizeRoles(self):
        '''This method converts, within self.permissions, every role to a
           Role instance. Every used role is stored in self.usedRoles.'''
        for permission, roles in self.permissions.iteritems():
            if isinstance(roles, basestring) or isinstance(roles, Role):
                role = self.getRole(roles)
                self.permissions[permission] = {role.name: role}
            elif isinstance(roles, dict):
                for name, role in roles.iteritems():
                    roles[name] = self.getRole(role)
            else:
                # "roles" is a list or tuple, or None (nobody may have this
                # permission).
                d = {}
                if roles is not None:
                    for role in roles:
                        role = self.getRole(role)
                        d[role.name] = role
                self.permissions[permission] = d

    def getUsedRoles(self): return self.usedRoles.values()

    def addRoles(self, roles, permissions=()):
        '''Adds p_roles in self.permissions. p_roles can be a role name, a Role
           instance or a list of names and/or Role instances. If p_permissions
           is specified, roles are added to those permissions only. Else, roles
           are added for every permission within self.permissions.'''
        # Standardize parameters
        if type(roles) not in sequenceTypes: roles = (roles,)
        if isinstance(permissions, basestring): permissions = (permissions,)
        for perm, existingRoles in self.permissions.iteritems():
            if permissions and (perm not in permissions): continue
            for role in roles:
                # Do nothing if "role" is already among existing roles
                name = isinstance(role, basestring) and role or role.name
                if name in existingRoles: continue
                # Add the role for this permission
                existingRoles[name] = self.getRole(role)

    def removeRoles(self, roleNames, permissions=()):
        '''Removes p_roleNames within dict self.permissions. If p_permissions is
           specified, removal is restricted to those permissions. Else, removal
           occurs throughout the whole dict self.permissions.'''
        if isinstance(roleNames, basestring): roleNames = (roleNames,)
        if isinstance(permissions, basestring): permissions = (permissions,)
        for perm, roles in self.permissions.iteritems():
            if permissions and (perm not in permissions): continue
            for name in roleNames:
                # Remove this role if present in roles for this permission
                if name in roles:
                    del roles[name]

    def setRoles(self, roleNames, permissions=()):
        '''Sets p_rolesNames for p_permissions if not empty, for every
           permission in self.permissions else.'''
        if isinstance(roleNames, basestring): roleNames = (roleNames,)
        if isinstance(permissions, basestring): permissions = (permissions,)
        for perm in self.permissions.iterkeys():
            if permissions and (perm not in permissions): continue
            roles = self.permissions[perm] = {}
            for name in roleNames:
                roles[name] = self.getRole(name)

    def replaceRole(self, oldRoleName, newRoleName, permissions=()):
        '''Replaces p_oldRoleName by p_newRoleName. If p_permissions is
           specified, the replacement is restricted to those permissions. Else,
           replacements apply to the whole dict self.permissions.'''
        if isinstance(permissions, basestring): permissions = (permissions,)
        for perm, roles in self.permissions.iteritems():
            if permissions and (perm not in permissions): continue
            # Find and replace p_oldRoleName with p_newRoleName
            if oldRoleName in roles:
                del roles[oldRoleName]
                roles[newRoleName] = self.getRole(newRoleName)

    def copyRoles(self, sourcePermission, destPermission):
        '''Overrides p_destPermission's roles with (a deep copy of)
           p_sourcePermission's roles.'''
        copiedRoles = copy.deepcopy(self.permissions[sourcePermission])
        self.permissions[destPermission] = copiedRoles

    def isIsolated(self, wf):
        '''Returns True if, from this state, we cannot reach another state. The
           workflow class is given in p_wf. Modifying a workflow for getting a
           state with auto-transitions only is a common technique for disabling
           a state in a workflow.'''
        if self.initial: return
        for tr in wf.__dict__.itervalues():
            if not isinstance(tr, Transition): continue
            # Ignore transitions that do not touch this state
            if not tr.hasState(self, True) and not tr.hasState(self, False):
                continue
            # Transition "tr" has this state as start or end state. If start and
            # end states are different, it means that the state is not
            # isolated.
            if tr.isSingle():
                for state in tr.states:
                    if state != self: return
            else:
                for start, end in tr.states:
                    # Bypass (start, end) pairs having nothing to do with self
                    if (start != self) and (end != self): continue
                    if (start != self) or (end != self): return
        # If we are here, either there was no transition starting from self,
        # either all transitions were auto-transitions: self is then isolated.
        return True

# ------------------------------------------------------------------------------
class Transition:
    '''Represents a workflow transition'''
    UNTRIGGERABLE = 'Transition "%s" on %s can\'t be triggered.'

    def __init__(self, states, condition=True, action=None, show=True,
                 confirm=False, group=None, icon=None, redirect=None,
                 historizeActionMessage=False):
        # In its simpler form, "states" is a list of 2 states:
        # (fromState, toState). But it can also be a list of several
        # (fromState, toState) sub-lists. This way, you may define only 1
        # transition at several places in the state-transition diagram. It may
        # be useful for "undo" transitions, for example.
        self.states = self.standardiseStates(states)
        self.condition = condition
        if isinstance(condition, basestring):
            # The condition specifies the name of a role
            self.condition = Role(condition)
        self.action = action
        self.show = show # If False, the end user will not be able to trigger
        # the transition. It will only be possible by code.
        self.confirm = confirm # If True, a confirm popup will show up.
        self.group = Group.get(group)
        # The user may specify a specific icon to show for this transition.
        self.icon = icon or 'transition'
        # If redirect is None, once the transition will be triggered, Appy will
        # perform an automatic redirect:
        # (a) if you were on some "view" page, Appy will redirect you to this
        #     page (thus refreshing it entirely);
        # (b) if you were in a list of objects, Appy will Ajax-refresh the row
        #     containing the object from which you triggered the transition.
        # Case (b) can be problematic if the transition modifies the list of
        # objects, or if it modifies other elements shown outside this list.
        # If you specify "redirect" being
        # ----------------------------------------------------------------------
        # "page"     | case (a) will always apply;
        # ----------------------------------------------------------------------
        # "referer"  | the entire page will be refreshed with the referer page.
        # ----------------------------------------------------------------------
        self.redirect = redirect
        # When a transition is triggered, the corresponding event is added in
        # the object's history. If p_historizeActionMessage is True, the message
        # returned by self.action (if any) will be appended to this event's
        # comment.
        self.historizeActionMessage = historizeActionMessage

    def standardiseStates(self, states):
        '''Get p_states as a list or a list of lists. Indeed, the user may also
           specify p_states as a tuple or tuple of tuples. Having lists allows
           us to easily perform changes in states if required.'''
        if isinstance(states[0], State):
            if isinstance(states, tuple): return list(states)
            return states
        return [[start, end] for start, end in states]

    def getName(self, wf):
        '''Returns the name for this state in workflow p_wf'''
        for name in dir(wf):
            value = getattr(wf, name)
            if (value == self): return name

    def getEndStateName(self, wf, startStateName=None):
        '''Returns the name of p_self's end state. If p_self is a
           multi-transition, the name of a specific p_startStateName can be
           given.'''
        if self.isSingle():
            return self.states[1].getName(wf)
        else:
            for start, end in self.states:
                if not startStateName:
                    return end.getName(wf)
                else:
                    if start.getName(wf) == startStateName:
                        return end.getName(wf)

    def getUsedRoles(self):
        '''self.condition can specify a role'''
        res = []
        if isinstance(self.condition, Role):
            res.append(self.condition)
        return res

    def isSingle(self):
        '''If this transition is only defined between 2 states, returns True.
           Else, returns False.'''
        return isinstance(self.states[0], State)

    def addAction(self, action):
        '''Adds an p_action in self.action'''
        actions = self.action
        if not actions:
            self.action = action
        elif isinstance(actions, list):
            actions.append(action)
        elif isinstance(actions, tuple):
            self.action = list(actions)
            self.action.append(action)
        else: # A single action is already defined
            self.action = [actions, action]

    def _replaceStateIn(self, oldState, newState, states):
        '''Replace p_oldState by p_newState in p_states.'''
        if oldState not in states: return
        i = states.index(oldState)
        del states[i]
        states.insert(i, newState)

    def replaceState(self, oldState, newState):
        '''Replace p_oldState by p_newState in self.states.'''
        if self.isSingle():
            self._replaceStateIn(oldState, newState, self.states)
        else:
            for i in range(len(self.states)):
                self._replaceStateIn(oldState, newState, self.states[i])

    def removeState(self, state):
        '''For a multi-state transition, this method removes every state pair
           containing p_state.'''
        if self.isSingle():
            raise WorkflowException('To use for multi-transitions only')
        i = len(self.states) - 1
        while i >= 0:
            if state in self.states[i]:
                del self.states[i]
            i -= 1
        # This transition may become a single-state-pair transition.
        if len(self.states) == 1:
            self.states = self.states[0]

    def setState(self, state):
        '''Configure this transition as being an auto-transition on p_state.
           This can be useful if, when changing a workflow, one wants to remove
           a state by isolating him from the rest of the state diagram and
           disable some transitions by making them auto-transitions of this
           disabled state.'''
        self.states = [state, state]

    def isShowable(self, workflow, obj):
        '''Is this transition showable?'''
        if callable(self.show):
            return self.show(workflow, obj.appy())
        else:
            return self.show

    def hasState(self, state, isFrom):
        '''If p_isFrom is True, this method returns True if p_state is a
           starting state for p_self. If p_isFrom is False, this method returns
           True if p_state is an ending state for p_self.'''
        stateIndex = 1
        if isFrom:
            stateIndex = 0
        if self.isSingle():
            res = state == self.states[stateIndex]
        else:
            res = False
            for states in self.states:
                if states[stateIndex] == state:
                    res = True
                    break
        return res

    def replaceRoleInCondition(self, old, new):
        '''When self.condition is a tuple or list, this method replaces role
           p_old by p_new. p_old and p_new can be strings or Role instances.'''
        condition = self.condition
        if isinstance(old, Role): old = old.name
        # Ensure we have a list
        if isinstance(condition, tuple): condition = list(condition)
        if not isinstance(condition, list):
            raise WorkflowException('m_replaceRoleInCondition can only be ' \
              'used if transition.condition is a sequence.')
        # Find the p_old role
        i = -1
        found = False
        for cond in condition:
            i += 1
            if isinstance(cond, Role): cond = cond.name
            if cond == old:
                found = True
                break
        if not found: return
        del condition[i]
        condition.insert(i, new)
        self.condition = tuple(condition)

    def isTriggerable(self, obj, wf, noSecurity=False):
        '''Can this transition be triggered on p_obj ?'''
        wf = wf.__instance__ # We need the prototypical instance here
        # Checks that the current state of the object is a start state for this
        # transition.
        objState = obj.State(name=False)
        if self.isSingle():
            if objState != self.states[0]: return False
        else:
            startFound = False
            for startState, stopState in self.states:
                if startState == objState:
                    startFound = True
                    break
            if not startFound: return False
        # Check that the condition is met, excepted if noSecurity is True
        if noSecurity: return True
        user = obj.getTool().getUser()
        if isinstance(self.condition, Role):
            # Condition is a role. Transition may be triggered if the user has
            # this role.
            return user.hasRole(self.condition.name, obj)
        elif callable(self.condition):
            return self.condition(wf, obj.appy())
        elif type(self.condition) in (tuple, list):
            # It is a list of roles and/or functions. Transition may be
            # triggered if user has at least one of those roles and if all
            # functions return True.
            hasRole = None
            for condition in self.condition:
                # "Unwrap" role names from Role instances
                if isinstance(condition, Role): condition = condition.name
                if isinstance(condition, basestring): # It is a role
                    if hasRole is None:
                        hasRole = False
                    if user.hasRole(condition, obj):
                        hasRole = True
                else: # It is a method
                    res = condition(wf, obj.appy())
                    if not res: return res # False or a No instance
            if hasRole != False:
                return True
        else:
            return bool(self.condition)

    def executeAction(self, obj, wf):
        '''Executes the action related to this transition'''
        msg = ''
        obj = obj.appy()
        wf = wf.__instance__ # We need the prototypical instance here
        if type(self.action) in (tuple, list):
            # We need to execute a list of actions
            for act in self.action:
                msgPart = act(wf, obj)
                if msgPart: msg += msgPart
        else: # We execute a single action only
            msgPart = self.action(wf, obj)
            if msgPart: msg += msgPart
        return msg

    def executeCommonAction(self, obj, name, wf, fromState):
        '''Executes the action that is common to any transition, named
           "onTrigger" on the workflow class by convention. The common action is
           executed before the transition-specific action (if any).'''
        obj = obj.appy()
        wf = wf.__instance__ # We need the prototypical instance here
        wf.onTrigger(obj, name, fromState)

    def trigger(self, name, obj, wf, comment, doAction=True, doHistory=True,
                doSay=True, reindex=True, noSecurity=False, data=None):
        '''This method triggers this transition (named p_name) on p_obj. If
           p_doAction is False, the action that must normally be executed after
           the transition has been triggered will not be executed. If
           p_doHistory is False, there will be no trace from this transition
           triggering in p_obj's history. If p_doSay is False, we consider
           the transition as being triggered programmatically, and no message is
           returned to the user. If p_reindex is False, object reindexing will
           be performed by the caller method. If p_data is specified, it is a
           dict containing custom data that will be integrated into the history
           event. Be careful: if keys in this dict correspond to standard event
           keys ("action", "time"...) they will override it.'''
        # "Triggerability" and security checks
        isInit = name == '_init_'
        if not isInit and not self.isTriggerable(obj, wf,noSecurity=noSecurity):
            raise WorkflowException(self.UNTRIGGERABLE % \
                                    (name, obj.absolute_url()))
        # Create the workflow_history dict if it does not exist
        if not hasattr(obj.aq_base, 'workflow_history'):
            from persistent.mapping import PersistentMapping
            obj.workflow_history = PersistentMapping()
        # Create the event list if it does not exist in the dict. The
        # overstructure (a dict with a key 'appy') is only there for historical
        # reasons and will change in Appy 1.0
        if not obj.workflow_history: obj.workflow_history['appy'] = ()
        # Identify the target state for this transition
        if self.isSingle():
            targetState = self.states[1]
            targetStateName = targetState.getName(wf)
        else:
            startState = obj.State(name=False)
            for sState, tState in self.states:
                if startState == sState:
                    targetState = tState
                    targetStateName = targetState.getName(wf)
                    break
        # Add the event in the object history
        action = name
        if isInit:
            action = None
            fromState = None
        else:
            fromState = obj.State() # Remember the "from" (=start) state
        if not doHistory:
            comment = '_invisible_'
            if data: data['actor'] = 'system'
            else:    data = {'actor': 'system'}
        if not data: data = emptyDict
        event = obj.addHistoryEvent(action, review_state=targetStateName,
                                    comments=comment, **data)
        # Execute the action that is common to all transitions, if defined
        if doAction and hasattr(wf, 'onTrigger'):
            self.executeCommonAction(obj, name, wf, fromState)
        # Execute the related action (if any)
        msg = ''
        if doAction and self.action: msg = self.executeAction(obj, wf)
        # Append the action message to the history event when relevant
        if doHistory and msg and self.historizeActionMessage:
            if not comment: comment = msg
            else: comment = '%s<br/><br/>%s' % (comment, msg)
            event['comments'] = comment
            obj.workflow_history['appy'] = obj.workflow_history['appy'][:-1] + \
                                           (event,)
        # Reindex the object if required. Not only security-related indexes
        # (Allowed, State) need to be updated here.
        if reindex and not isInit and not obj.isTemporary(): obj.reindex()
        # Return a message to the user if needed
        if not doSay: return
        if not msg: msg = obj.translate('object_saved')
        return msg

    def onUiRequest(self, obj, wf, name, req):
        '''Executed when a user wants to trigger this transition from the UI'''
        tool = obj.getTool()
        # Trigger the transition
        msg = self.trigger(name, obj, wf, req.get('popupComment', ''),
                           reindex=False)
        # Reindex obj if required
        if not obj.isTemporary(): obj.reindex()
        # If we are called from an Ajax request, simply return msg
        if hasattr(req, 'pxContext') and req.pxContext['ajax']: return msg
        # If we are viewing the object and if the logged user looses the
        # permission to view it, redirect the user to its home page.
        if msg: obj.say(msg)
        # Return to obj/view, excepted if the object is not viewable anymore
        if obj.mayView():
            if self.redirect == 'referer':
                back = req['HTTP_REFERER']
            else:
                back = obj.getUrl(nav=req.get('nav', 'no'),
                                  page=req.get('page', 'main'),
                                  inPopup=req.get('popup', '0') == '1')
        else:
            back = tool.getHomePage()
        tool.goto(back)

    @staticmethod
    def getBack(workflow, transition):
        '''Returns the name of the transition (in p_workflow) that "cancels" the
           triggering of p_transition and allows to go back to p_transition's
           start state.'''
        # Get the end state(s) of p_transition
        transition = getattr(workflow, transition)
        # Browse all transitions and find the one starting at p_transition's end
        # state and coming back to p_transition's start state.
        for trName, tr in workflow.__dict__.iteritems():
            if not isinstance(tr, Transition) or (tr == transition): continue
            if transition.isSingle():
                if tr.hasState(transition.states[1], True) and \
                   tr.hasState(transition.states[0], False): return trName
            else:
                startOk = False
                endOk = False
                for start, end in transition.states:
                    if (not startOk) and tr.hasState(end, True):
                        startOk = True
                    if (not endOk) and tr.hasState(start, False):
                        endOk = True
                    if startOk and endOk: return trName

class UiTransition:
    '''Represents a widget that displays a transition'''
    pxView = Px('''
     <x var="label=transition.title;
             inButtons=layoutType == 'buttons';
             css=ztool.getButtonCss(label, inButtons)">

      <!-- Real button -->
      <input if="transition.mayTrigger" type="button" class=":css"
             var="back=transition.getBackHook(zobj, inButtons, q, backHook)"
             name=":transition.name" style=":url(transition.icon, bg=True)"
             value=":label"
             onclick=":'triggerTransition(%s,this.name,%s,%s)' % \
                        (q(formId), q(transition.confirm), back)"/>

      <!-- Fake button, explaining why the transition can't be triggered -->
      <input if="not transition.mayTrigger" type="button"
             class=":'fake %s' % css" style=":url('fake', bg=True)"
             value=":label" title=":transition.reason"/></x>''')

    def __init__(self, name, transition, obj, mayTrigger):
        self.name = name
        self.transition = transition
        self.type = 'transition'
        self.icon = transition.icon
        label = obj.getWorkflowLabel(name)
        self.title = obj.translate(label)
        if transition.confirm:
            msg = obj.translate('%s_confirm' % label, blankOnError=True) or \
                  obj.translate('action_confirm')
            self.confirm = msg
        else:
            self.confirm = ''
        # May this transition be triggered via the UI?
        self.mayTrigger = True
        self.reason = ''
        if not mayTrigger:
            self.mayTrigger = False
            self.reason = mayTrigger.msg
        # Required by the UiGroup
        self.colspan = 1

    def getBackHook(self, obj, inButtons, q, backHook=None):
        '''If, when the transition has been triggered, we must ajax-refresh some
           part of the page, this method will return the ID of the corresponding
           DOM node. Else (ie, the entire page needs to be refreshed), it
           returns None.'''
        if inButtons and not self.transition.redirect:
            return q(backHook or obj.id)
        return 'null'

# ------------------------------------------------------------------------------
class Permission:
    '''If you need to define a specific read or write permission for some field
       on a gen-class, you use the specific boolean attrs
       "specificReadPermission" or "specificWritePermission". When you want to
       refer to those specific read or write permissions when
       defining a workflow, for example, you need to use instances of
       "ReadPermission" and "WritePermission", the 2 children classes of this
       class. For example, if you need to refer to write permission of
       attribute "t1" of class A, write: WritePermission("A.t1") or
       WritePermission("x.y.A.t1") if class A is not in the same module as
       where you instantiate the class.

       Note that this holds only if you use attributes "specificReadPermission"
       and "specificWritePermission" as booleans. When defining named
       (string) permissions, for referring to it you simply use those strings,
       you do not create instances of ReadPermission or WritePermission.'''

    def __init__(self, fieldDescriptor):
        self.fieldDescriptor = fieldDescriptor

    def getName(self, wf, appName):
        '''Returns the name of this permission'''
        className, fieldName = self.fieldDescriptor.rsplit('.', 1)
        if className.find('.') == -1:
            # The related class resides in the same module as the workflow
            fullClassName= '%s_%s' % (wf.__module__.replace('.', '_'),className)
        else:
            # className contains the full package name of the class
            fullClassName = className.replace('.', '_')
        # Read or Write ?
        if self.__class__.__name__ == 'ReadPermission': access = 'Read'
        else: access = 'Write'
        return '%s: %s %s %s' % (appName, access, fullClassName, fieldName)

class ReadPermission(Permission): pass
class WritePermission(Permission): pass

# Standard workflows -----------------------------------------------------------
class WorkflowAnonymous:
    '''One-state workflow allowing anyone to consult and Manager to edit'''
    ma = 'Manager'
    o = 'Owner'
    everyone = (ma, 'Anonymous', 'Authenticated')
    active = State({r:everyone, w:(ma, o), d:(ma, o)}, initial=True)

class WorkflowAuthenticated:
    '''One-state workflow allowing authenticated users to consult and Manager
       to edit.'''
    ma = 'Manager'
    o = 'Owner'
    authenticated = (ma, 'Authenticated')
    active = State({r:authenticated, w:(ma, o), d:(ma, o)}, initial=True)

class WorkflowOwner:
    '''Workflow allowing only manager and owner to consult and edit'''
    ma = 'Manager'
    o = 'Owner'
    # States
    active = State({r:(ma, o), w:(ma, o), d:ma}, initial=True)
    inactive = State({r:(ma, o), w:ma, d:ma})
    # Transitions
    def doDeactivate(self, obj):
        '''Prevent user "admin" from being deactivated'''
        if isinstance(obj, User) and (obj.login == 'admin'):
            raise WorkflowException('Cannot deactivate admin.')
    deactivate = Transition( (active, inactive), condition=ma,
                             action=doDeactivate)
    reactivate = Transition( (inactive, active), condition=ma)
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from base64 import encodebytes
from mimetypes import guess_type
import xml.sax, http.client, urllib, urllib.parse, ssl
import os, sys, re, stat, time, socket, random, hashlib, gzip

from appy.utils import copyData
from appy.model.utils import Object
from appy.utils.string import Normalize
from appy.xml.marshaller import Marshaller
from appy.xml.unmarshaller import Unmarshaller

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class FormDataEncoder:
    '''Allows to encode form data for sending it through a HTTP request'''
    def __init__(self, data):
        self.data = data # The data to encode, as a dict

    def marshalValue(self, name, value):
        if isinstance(value, str):
            return '%s=%s' % (name, urllib.parse.quote(value))
        elif isinstance(value, float):
            return '%s:float=%s' % (name, value)
        elif isinstance(value, int):
            return '%s:int=%s' % (name, value)
        elif isinstance(value, long):
            res = '%s:long=%s' % (name, value)
            if res[-1] == 'L':
                res = res[:-1]
            return res
        else:
            raise Exception('Cannot encode value %s' % str(value))

    def encode(self):
        res = []
        for name, value in self.data.items():
            res.append(self.marshalValue(name, value).encode())
        return b'&'.join(res)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class SoapDataEncoder:
    '''Allows to encode SOAP data for sending it through a HTTP request'''
    namespaces = {'s'  : 'http://schemas.xmlsoap.org/soap/envelope/',
                  'xsd': 'http://www.w3.org/2001/XMLSchema',
                  'xsi': 'http://www.w3.org/2001/XMLSchema-instance'}
    namespacedTags = {'Envelope': 's', 'Body': 's', '*': 'py'}

    def __init__(self, data, namespace='http://appyframework.org'):
        self.data = data
        # p_data can be:
        # - a string already containing a complete SOAP message
        # - a Python object, that we will convert to a SOAP message
        # Define the namespaces for this request
        self.ns = self.namespaces.copy()
        self.ns['py'] = namespace

    def encode(self):
        # Do nothing if we have a SOAP message already
        if isinstance(self.data, str): return self.data
        # self.data is here a Python object. Wrap it in a SOAP Body.
        soap = Object(Body=self.data)
        # Marshall it
        marshaller = Marshaller(rootTag='Envelope', namespaces=self.ns,
                                namespacedTags=self.namespacedTags)
        return marshaller.marshall(soap)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class JsonDecoder:
    '''Converts JSON data into Pyton data structures'''
    # Boolean and None values are different in JSON and Python
    context = {'true': True, 'false': False, 'null': None}
    UNREADABLE = 'Unreadable JSON string: %s'

    @classmethod
    def convertDict(class_, d):
        '''Returns a appy.Object instance representing dict p_d'''
        res = Object()
        for name, value in d.items():
            # Ensure "name" will be a valid attribute name for a Python object
            n = Normalize.alphanum(name, keepUnderscore=True)
            setattr(res, n, class_.convertValue(value))
        return res

    @classmethod
    def convertList(class_, l):
        '''Every item being a dict in p_l is converted to an object'''
        i = len(l) - 1
        while i >= 0:
            l[i] = class_.convertValue(l[i])
            i -= 1

    @classmethod
    def convertValue(class_, val):
        '''Converts a JSON p_val into a Python value'''
        if isinstance(val, list):
            class_.convertList(val)
            res = val
        elif isinstance(val, dict):
            res = class_.convertDict(val)
        else:
            # In all other cases, no conversion is needed
            res = val
        return res

    @classmethod
    def decode(class_, jsonData):
        '''Converts JSON data received in a string (p_jsonData) to a Python data
           structure. JSON dicts are converted to Python objects.'''
        try:
            return class_.convertValue(eval(jsonData, class_.context))
        except SyntaxError:
            # The presence of char "\r" may pose problem
            jsonData = jsonData.replace('\r', '')
            try:
                return class_.convertValue(eval(jsonData, class_.context))
            except SyntaxError as err:
                raise ResourceError(JsonDecoder.UNREADABLE % jsonData)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class DigestRealm:
    '''Represents information delivered by a server requiring Digest-based
       HTTP authentication.'''
    rex = re.compile('(\w+)="(.*?)"')
    # List of attributes whose values must not be quoted
    unquoted = ('algorithm', 'qop', 'nc')

    def md5(self, value):
        '''Produces the MD5 message digest for p_value, as a hexadecimal
           32-chars string.'''
        return hashlib.md5(str(value).encode()).hexdigest()

    def __init__(self, info):
        '''p_info is the content of header key WWW-Authenticate. This
           constructor parses it and unwraps data into self's attributes.'''
        for name, value in self.rex.findall(info):
            # For attribute "qpop", split values into a list
            if name == 'qop':
                value = value.split(',')
                for i in range(len(value)):
                    value[i] = value[i].strip()
            setattr(self, name, value)
        # Set some default values
        if not hasattr(self, 'algorithm'):
            self.algorithm = 'MD5'

    def buildCredentials(self, resource, uri, httpMethod='GET'):
        '''Builds credentials to transmit to the server'''
        login = resource.username
        realm = self.realm
        algorithm = self.algorithm
        nonce = self.nonce
        # Get the "path" part of the URI
        parsed = urllib.parse.urlparse(uri).path
        # Compute a client random nouce
        cnonce = self.md5(random.random())
        # Collect credentials info in a dict
        res = {'username': login, 'uri': uri, 'realm': realm, 'nonce': nonce,
               'algorithm': algorithm}
        # Add optional attribute "opaque"
        if hasattr(self, 'opaque'): res['opaque'] = self.opaque
        # Precompute the "HA1" part of the response, that depends on the
        # algorithm in use (MD5 or MD5-sess).
        ha1 = self.md5('%s:%s:%s' % (login, realm, resource.password))
        if algorithm == 'MD5-sess':
            ha1 = self.md5('%s:%s:%s' % (ha1, nonce, cnonce))
        # Take into account the quality of protection (qop)
        hasQop = hasattr(self, 'qop')
        if hasQop:
            qop = res['qop'] = self.qop[0]
            res['cnonce'] = cnonce
            res['nc'] = '00000001'
        else:
            qop = 'auth'
        # Precompute the "HA2" part of the response, that depends on qop
        if qop == 'auth-int':
            entity = self.md5('entity')
            ha2 = self.md5('%s:%s:%s' % (httpMethod, uri, entity))
        else:
            ha2 = self.md5('%s:%s' % (httpMethod, uri))
        # Compute the complete response
        if hasQop:
            response = self.md5('%s:%s:%s:%s:%s:%s' % \
                                (ha1, nonce, res['nc'], cnonce, qop, ha2))
        else:
            response = self.md5('%s:%s:%s' % (ha1, nonce, ha2))
        res['response'] = response
        # Convert the dict to a formatted list, quoting values when relevant
        attrs = []
        for name, value in res.items():
            if name not in self.unquoted:
                value = '"%s"' % value
            attrs.append('%s=%s' % (name, value))
        # Produce the final value
        return 'Digest %s' % ', '.join(attrs)

    def __repr__(self):
        pairs = ['%s=%s' % (k, v) for k, v in self.__dict__.items()]
        return '<Realm %s>' % ','.join(pairs)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class HttpResponse:
    '''Stores information about a HTTP response'''
    def __init__(self, resource, response, body, duration=None, utf8=True,
                 responseType=None):
        self.resource = resource
        self.code = response.status # The return code, ie 404, 200, 500...
        self.text = response.reason # Textual description of the code
        self.headers = response.msg # A dict-like object containing the headers
        self.body = body # The body of the HTTP response
        # p_duration, if given, is the time, in seconds, we have waited, before
        # getting this response after having sent the request.
        self.duration = duration
        self.utf8 = utf8
        # The following attribute may contain specific data extracted from
        # the previous fields. For example, when response if 302 (Redirect),
        # self.data contains the URI where we must redirect the user to.
        self.data = self.extractData(responseType)
        self.response = response

    def __repr__(self, complete=False):
        duration = suffix = ''
        if self.duration: duration = ', got in %.4f seconds' % self.duration
        if complete:
            suffix = '\n*Headers* %s\n*Body* %s\n' %  \
                     (str(self.headers), str(self.body))
        return '<HttpResponse %s (%s)%s%s>' % \
               (self.code, self.text, duration, suffix)
    def get(self): return self.__repr__(complete=True)

    def getResponseUrl(self, url):
        '''Get the URL path'''
        parts = urllib.parse.urlparse(url)
        r = parts.path or '/'
        if parts.query:
            r += '?%s' % parts.query
        return r

    def extractContentType(self, contentType):
        '''Extract the content type from the HTTP header, potentially removing
           encoding-related data.'''
        i = contentType.find(';')
        if i != -1: return contentType[:i]
        return contentType

    def extractCookies(self, headers):
        '''Extract received cookies and put them in self.resource.headers'''
        if 'Set-Cookie' not in headers: return
        # Several "Set-Cookie" can be returned by the server, so do not access
        # the dict-like p_headers, into which only the first "Set-Cookie" will
        # be present: access the underlying p_headers._headers tuple.
        for key, value in headers._headers:
            if key != 'Set-Cookie': continue
            cookie = value.split(';')[0]
            name, value = cookie.split('=', 1)
            # Delete the cookie if it has expired
            if value.strip('"') == 'deleted':
                # Do not set this cookie, and remove it if already present
                if name in self.resource.cookies:
                    del(self.resource.cookies[name])
            else:
                self.resource.cookies[name] = value

    xmlHeaders = ('text/xml', 'application/xml', 'application/soap+xml')
    def extractData(self, responseType=None):
        '''This method extracts, from the various parts of the HTTP response,
           some useful information:
           * it will find the URI where to redirect the user to if self.code
             is 302;
           * it will return authentication-related data, if present, if
             self.code is 401;
           * it will unmarshall XML or JSON data into Python objects;
           * ...'''
        # Extract information from HTTP headers when relevant
        headers = self.headers
        self.extractCookies(headers)
        if self.code == 302:
            # The URL to redirect to is in header key 'location'
            return self.getResponseUrl(headers['location'])
        elif self.code == 401:
            authInfo = headers.get('WWW-Authenticate')
            return authInfo and DigestRealm(authInfo) or None
        # Determine the response type from the HTTP response, or, if not found,
        # use p_responseType that may have been given.
        responseType = headers.get('Content-Type')
        if not responseType: return
        # Apply some transform on the response content depending on its type
        contentType = self.extractContentType(responseType)
        # Manage JSON content
        if contentType == 'application/json':
            return JsonDecoder.decode(self.body)
        # Manage XML content
        for xmlHeader in self.xmlHeaders:
            # Ensure this is XML
            if not contentType.startswith(xmlHeader): continue
            # Return an unmarshalled version of the XML content, for easy use
            try:
                parser = Unmarshaller()
                res = parser.parse(self.body)
                if parser.rootTag == 'exception':
                    # This is an exception: "res" contains the traceback
                    raise Resource.Error('Distant server exception: %s' % res)
                return res
            except xml.sax.SAXParseException as se:
                raise Resource.Error('Invalid XML response (%s)' % str(se))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
urlRex = re.compile(r'(http[s]?)://([^:/]+)(:[0-9]+)?(/.+)?', re.I)
binaryRex = re.compile(r'[\000-\006\177-\277]')

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Resource:
    '''Every instance of this class represents some web resource accessible
       through HTTP.'''
    class Error(Exception): pass
    TIMEOUT_ERROR = 'Timed out after %d second(s).'
    standardPorts = {'http': 80, 'https': 443}

    def __init__(self, url, username=None, password=None, measure=False,
                 utf8=True, authMethod='Basic', timeout=10):
        self.username = username
        self.password = password
        self.authMethod = authMethod
        self.url = url
        self.utf8 = utf8
        # A timeout (in seconds) used when sending blocking requests to the
        # resource.
        self.timeout = timeout
        # If p_measure is True, we will measure, for every request sent, the
        # time we wait until we receive the response.
        self.measure = measure
        # If measure is True, we will store hereafter, the total time (in
        # seconds) spent waiting for the server for all requests sent through
        # this resource object.
        self.serverTime = 0
        # Split the URL into its components
        res = urlRex.match(url)
        if res:
            protocol, host, port, uri = res.groups()
            self.protocol = protocol
            self.host = host
            if port:
                self.port = int(port[1:])
            else:
                self.port = Resource.standardPorts[self.protocol]
            self.uri = uri or '/'
        else: raise self.Error('Wrong URL: %s' % str(url))
        # If some headers must be sent with any request sent through this
        # resource, you can store them in the following dict.
        self.headers = {'Host': self.getHeaderHost(), 'User-Agent': 'Appy',
                        'Connection': 'close', 'Accept': '*/*',
                        'Accept-Encoding': 'gzip, identity'}
        # Cookies defined hereafter will be included in self.headers at every
        # request.
        self.cookies = {}

    def getHeaderHost(self):
        '''Gets the content of header key "Host"'''
        # Insert the port number if not standard
        suffix = '' if self.port == Resource.standardPorts[self.protocol] \
                    else ':%d' % self.port
        return '%s%s' % (self.host, suffix)

    def __repr__(self):
        return '<Resource at %s>' % self.url

    def completeHeaders(self, headers):
        # Get standard header values from self.headers if not defined in
        # p_headers
        if self.headers:
            for k, v in self.headers.items():
                if k not in headers:
                    headers[k] = v
        # Add cookies
        if self.cookies:
            headers['Cookie'] = '; '.join(['%s=%s' % (k, v) \
                                          for k, v in self.cookies.items()])
        # Add credentials-related headers when relevant
        if not (self.username and self.password): return
        if 'Authorization' in headers: return
        if self.authMethod == 'Basic':
            creds = ('%s:%s' % (self.username, self.password)).encode()
            authorization = '%s %s' % (self.authMethod,
                                       encodebytes(creds).strip().decode())
            headers['Authorization'] = authorization

    def readResponse(self, response):
        '''Reads the response content. Unzip the result when appropriate'''
        headers = response.headers
        if headers.get('Content-Encoding') == 'gzip':
            # Unzip the content
            f = gzip.GzipFile(fileobj=response)
        else:
            f = response
        # Read response bytes
        r = f.read()
        # Decode the response when relevant
        contentType = headers.get('Content-Type')
        if contentType and contentType.startswith('image/'):
            pass # No need to decode it
        else:
            try:
                r = r.decode()
            except UnicodeDecodeError:
                pass
        return r

    def send(self, method, uri, body=None, headers=None, bodyType=None,
             responseType=None):
        '''Sends a HTTP request with p_method, for p_uri'''
        hc = http.client
        if self.protocol == 'https':
            context = ssl.SSLContext()
            context.verify_mode = ssl.CERT_NONE
            conn = hc.HTTPSConnection(self.host, self.port,
                                      timeout=self.timeout, context=context)
        else:
            conn = hc.HTTPConnection(self.host, self.port, timeout=self.timeout)
        try:
            conn.connect()
        except socket.gaierror as sge:
            raise self.Error('Check your Internet connection (%s)' % str(sge))
        except socket.timeout as se:
            raise self.Error(self.TIMEOUT_ERROR % conn.timeout)
        except socket.error as se:
            raise self.Error('Connection error (%s)' % str(se))
        # Tell what kind of HTTP request it will be
        conn.putrequest(method, uri, skip_host=True)
        # Add HTTP headers
        if headers is None: headers = {}
        self.completeHeaders(headers)
        for k, v in headers.items(): conn.putheader(k, v)
        conn.endheaders()
        # Add HTTP body
        if body:
            if not bodyType: bodyType = 'string'
            copyData(body, conn, 'send', type=bodyType)
        # Send the request, get the reply
        if self.measure: startTime = time.time()
        try:
            response = conn.getresponse()
        except socket.timeout as te:
            raise self.Error(self.TIMEOUT_ERROR % conn.timeout)
        if self.measure: endTime = time.time()
        body = self.readResponse(response)
        conn.close()
        # Return a smart object containing the various parts of the response
        duration = None
        if self.measure:
            duration = endTime - startTime
            self.serverTime += duration
        return HttpResponse(self, response, body, duration=duration,
                            utf8=self.utf8, responseType=responseType)

    def mkdir(self, name):
        '''Creates a folder named p_name in this resource.'''
        folderUri = self.uri + '/' + name
        #body = '<d:propertyupdate xmlns:d="DAV:"><d:set><d:prop>' \
        #       '<d:displayname>%s</d:displayname></d:prop></d:set>' \
        #       '</d:propertyupdate>' % name
        return self.send('MKCOL', folderUri)

    def delete(self, name):
        '''Deletes a file or a folder (and all contained files if any) named
           p_name within this resource.'''
        toDeleteUri = self.uri + '/' + name
        return self.send('DELETE', toDeleteUri)

    def add(self, path, name=''):
        '''Adds a file in this resource, whose p_path is passed'''
        size = os.stat(path)[stat.ST_SIZE]
        body = open(path, 'rb')
        name = os.path.basename(path)
        fileType, encoding = guess_type(path)
        fileUri = self.uri + '/' + name
        headers = {'Content-Length': str(size)}
        if fileType: headers['Content-Type'] = fileType
        if encoding: headers['Content-Encoding'] = encoding
        r = self.send('PUT', fileUri, body, headers, bodyType='file')
        body.close()
        return r

    def get(self, uri=None, headers=None, params=None, followRedirect=True,
            responseType=None):
        '''Perform a HTTP GET on the server. Parameters can be given as a dict
           in p_params. p_responseType will be used if no "Content-Type" key is
           found on the HTTP response.'''
        if not uri: uri = self.uri
        # Encode and append params if given
        if params:
            sep = ('?' in uri) and '&' or '?'
            uri = '%s%s%s' % (uri, sep, urllib.urlencode(params))
        r = self.send('GET', uri, headers=headers, responseType=responseType)
        # Follow redirect when relevant
        if (r.code == 302) and followRedirect:
            # Addition durations when "measure" is True
            duration = r.duration
            r = self.get(r.data, headers=headers)
            if self.measure: r.duration += duration
            return r
        # Perform Digest-based authentication when relevant
        if (r.code == 401) and (self.authMethod == 'Digest') and r.data:
            # Re-trigger the request with the correct authentication headers
            headers = headers or {}
            headers['Authorization'] = r.data.buildCredentials(self, uri)
            return self.get(uri=uri, headers=headers, params=params,
                       followRedirect=followRedirect, responseType=responseType)
        return r
    rss = get

    def post(self, data=None, uri=None, headers=None, encode='form',
             followRedirect=True):
        '''Perform a HTTP POST on the server. If p_encode is "form", p_data is
           considered to be a dict representing form data that will be
           form-encoded. Else, p_data will be considered as the ready-to-send
           body of the HTTP request.'''
        if not uri: uri = self.uri
        if headers is None: headers = {}
        # Prepare the data to send
        if encode == 'form':
            # Format the form data and prepare headers
            body = FormDataEncoder(data).encode()
            headers['Content-Type'] = 'application/x-www-form-urlencoded'
        else:
            body = data
        headers['Content-Length'] = str(len(body))
        r = self.send('POST', uri, headers=headers, body=body)
        if (r.code == 302) and followRedirect:
            # Update headers
            for key in ('Content-Type', 'Content-Length'):
                if key in headers: del headers[key]
            # Addition durations when "measure" is True
            duration = r.duration
            r = self.get(r.data, headers=headers)
            if self.measure: r.duration += duration
        return r

    def soap(self, data, uri=None, headers=None, namespace=None,
             soapAction=None):
        '''Sends a SOAP message to this resource. p_namespace is the URL of the
           server-specific namespace. If header value "SOAPAction" is different
           from self.url, specify it in p_soapAction.'''
        if not uri: uri = self.uri
        # Prepare the data to send
        data = SoapDataEncoder(data, namespace).encode()
        if headers is None: headers = {}
        headers['SOAPAction'] = soapAction or self.url
        # Content-type could be 'text/xml'
        headers['Content-Type'] = 'application/soap+xml;charset=UTF-8'
        res = self.post(data, uri, headers=headers, encode=None)
        # Unwrap content from the SOAP envelope
        if hasattr(res.data, 'Body'):
            res.data = res.data.Body
        return res
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



import os, sys
from distutils.core import setup
def findPackages(base):
    r = []
    for dir, dns, fns in os.walk(base + os.sep + 'appy'):
        r.append(dir[4:].replace('/', '.'))
    return r

# Python 2 or 3 ?
base = 'py%d' % sys.version_info[0]
if base == 'py3':
    dependencies = ['zodb', 'DateTime']
    python = '>=3.6'
else:
    dependencies = []
    python = '>=2.4'

setup(name = "appy", version = "1.0.7",
      description = "The Appy framework",
      long_description = "Appy is the simpliest way to build complex webapps.",
      author = "Gaetan Delannay",
      author_email = "gaetan.delannay@geezteem.com",
      license = "GPL", platforms="all",
      url = 'http://appyframework.org',
      packages = findPackages(base),
      package_dir = {'appy': base + '/appy'},
      package_data = {'':["*.*"]},
      install_requires = dependencies, python_requires = python)



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import os, sys, re, urllib.parse, logging, time, random
from appy.utils.client import Resource

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Configuration options for executing performance tests on a Appy server'''
    LOG_TARGET_ERROR = 'If multiple clients are executed, stdout cannot be ' \
      'used as log target. Specify the name of a file to log to.'

    def __init__(self, url, login, password,
                 scenario=[], clients=1, logTarget='stdout', silent=False,
                 alternateUrls=[]):
        # The Appy server URL(s) (including port)
        self.urls = [url]
        # The credentials used to authenticate to this server
        self.login = login
        self.password = password
        # The scenario to apply. Every entry can be a (relative) URL (starting
        # with a "/", the root of the site being "/") as a string or a Hit
        # instance (see below).
        self.scenario = scenario
        # The number of processes to fork for simulating simultaneous users.
        # Every client will execute the same scenario as defined above.
        self.clients = clients
        # Where to log test output. Can be 'stdout' or the name of a file. When
        # clients is > 1, output can't be 'stdout': it must be a thread-safe
        # file (that we manage via the standard logging module).
        self.logTarget = logTarget
        # If "silent" is True, a single log entry will be dumped by the script,
        # allowing for minimalistic log when multiple processes are run.
        self.silent = silent
        # You can specify a list of alternate URLs allowing to hit the same
        # resource as the one specified by p_url. In that case, every client
        # will choose one URL among [p_url] + alternateUrls for contacting the
        # server.
        if alternateUrls:
            self.urls += alternateUrls

    def check(self):
        '''Returns an error message if this config is not correct'''
        # stdout cannot be used when multiple clients are run
        if (self.clients > 1) and self.logTarget == 'stdout':
            return self.LOG_TARGET_ERROR

    def getUrl(self, i=0):
        '''Get the server URL to be used by the p_ith client'''
        urlsCount = len(self.urls)
        # If there is a single URL, return it
        if urlsCount == 1: return self.urls[0]
        # Compute the index, within self.urls, of the URL to return
        if i >= urlsCount:
            j = i
            while j >= urlsCount:
                j -= urlsCount
        else:
            j = i
        return self.urls[j]

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Hit:
    '''Represents a HTTP request on the server.
       If your scenario simply consists in a series of HTTP GET requests on the
       server, you do not need to use a Hit instance: simply define URLs as
       strings within Config.scenario. When things get more complex, use a Hit
       instance:
        * if you need to define a POST request;
        * if you need to define a URL by parsing the last (or a previous)
          request and finding your URL among it.
    '''
    # Regular expression for retrieving all "A" tags in a web page
    aRex = re.compile(b'<a\s+.*?href="(.*?)".*?>(.*?)</a>', re.S)

    def __init__(self, url, indexes=None, method='GET', comment=None,
                 id=None, baseHit=None):
        # The URL to hit on the tested server. If p_indexes is not None, this
        # URL is a regular expression that wil be applied to all "A" tags from
        # the last server response, to retrieve the URL to hit. Among all URLs
        # matching the regular expressions, those found at p_indexes will be
        # chosen.
        self.url = url
        self.indexes = indexes
        # GET or POST
        self.method = method
        # Some comment about this hit, that will be shown in the test report
        self.comment = comment
        # An identifier for this hit. This way, another hit will be able to
        # refer to this one
        self.id = id
        # The ID of the hit from which the URL for this one will be found. If
        # None, it means that the previous hit will be the base hit.
        self.baseHit = baseHit

    def cleanUrl(self, url):
        '''Removes any non-significative part in p_url'''
        parts = urllib.parse.urlparse(url)
        r = parts.path or '/'
        if parts.query:
            r += '?%s' % parts.query
        return r

    def sendOne(self, url, tester, comment='...', id=None):
        '''Called by m_send below to send a single request to the server. If an
           p_id(entifier) is given, it will be used to store the list of URLs
           found in the response body in tester.responseUrls. Returns a tuple
           (success, response).'''
        log = tester.log
        url = self.cleanUrl(url.decode())
        log('%s >> %s' % (url, comment))
        response = tester.server.get(url)
        log(response, 2)
        tester.countTime(response)
        if response.code != 200:
            log(' Error %d received.' % response.code)
            return False, response
        elif id:
            tester.responseUrls[id] = self.aRex.findall(response.body)
        return True, response

    def send(self, tester, previous):
        '''Send the request(s) corresponding to this hit to the tested server
           and returns a tuple (success, response) containing the result of
           executing the last request.'''
        log = tester.log
        if self.indexes:
            # Retrieve the URLs to get from the previous hit or a base hit
            if self.baseHit:
                if self.baseHit not in tester.responseUrls:
                    log('Hit "%s" was not found.' % self.baseHit, 2)
                    return False, None
                # Get the list of URLs retrieved by the base hit
                urls = tester.responseUrls[self.baseHit]
            else:
                # Get all "A" tags within previous body
                urls = self.aRex.findall(previous.body)
                if not urls:
                    log('No URL found in previous request.', 2)
                    return False, None
            # Apply regular expression "self.url" to match any URL from "urls"
            matched = []
            pattern = self.url
            for url, title in urls:
                if re.search(pattern, url):
                    matched.append((url, title))
            if not matched:
                log('Pattern "%s" not found among base hit.' % pattern, 2)
                return False, None
            # Get the URLs whose indexes are in self.indexes
            for i in self.indexes:
                # Check if this index is valid
                if i >= len(matched):
                    log('Wrong index %d (only %d URL(s) matched)' %
                        (i, len(matched)), 2)
                    return False, None
                # Build an ID corresponding to this index if this hit is
                # identified.
                id = '%s_%d' % (self.id, i) if self.id else None
                url, comment = matched[i]
                comment = comment.decode()
                success, response = self.sendOne(url, tester, comment, id)
                if not success: return False, None
            return True, response
        else:
            url = self.url
            comment = self.comment or '...'
            return self.sendOne(url, tester, comment, self.id)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Tester:
    '''Allows to execute performance tests on a Appy server by sending him a
       sequence of HTTP requests.'''

    def __init__(self, config):
        self.config = config
        self.server = None # Will be instantiated by m_runOne
        # The total time we have waited for server responses (seconds)
        self.duration = 0.0
        # Remember the list of URLs retrieved by every (identified) hit
        self.responseUrls = {} # ~{s_hitId: [(s_url, s_title)]}~
        if config.logTarget != 'stdout':
            logging.basicConfig(filename=config.logTarget, level=logging.DEBUG,
                                format='%(asctime)s: %(message)s')
        # These booleans indicate when to log to stdout and/or to the log file
        self.mustPrint = config.clients == 1
        self.mustLog = config.logTarget != 'stdout'

    def log(self, msg, blanks=0, level='info', force=False):
        '''Outputs on stdout and/or in a log file. p_blanks is the number of
           blanks that must be dumped at the start of the log entry.'''
        # Do not dump anything in silent mode, excepted if forced
        if self.config.silent and not force: return
        # Create the message to log and/or print
        msg = '%s%s' % (' '*blanks, msg)
        if self.config.clients > 1:
            msg = 'pid %d: %s' % (os.getpid(), msg)
        if self.mustPrint:
            print(msg)
        if self.mustLog:
            getattr(logging, level)(msg)

    def countTime(self, response):
        '''Count the time spent waiting for the server p_response'''
        self.duration += response.duration

    def sleep(self):
        '''Sleeps for a while. It is used at some early places in the scenario,
           to avoid processes to hit the server exactly at the same time.'''
        duration = random.random()
        time.sleep(duration)

    def authenticate(self):
        '''Logs into the p_server and returns a tuple (success, response):
           * "success"  is a boolean being True on sucessfull authentication;
           * "response" is the last HTTP response received by the server.
        '''
        server = self.server
        config = self.config
        log = self.log
        # Get the home page (so we know if the site is alive)
        self.sleep()
        log('*** Contacting %s as user "%s"...' % (server.url, config.login))
        log("/ >> site's home page... ")
        try:
            response = server.get()
            self.countTime(response)
        except Resource.Error as err:
            log('%s is unreachable (%s).' % (server.url, str(err)), 2)
            return False, None
        if response.code != 200:
            log('%s: got error %s.' % (server.url, response.code), 2)
            return False, None
        log(response, 2)
        # Define the parameters for the login
        data = {'js_enabled': 0, 'cookies_enabled': '', 'login_name': '',
                'pwd_empty': 0, '__ac_name': config.login,
                '__ac_password': config.password, 'submit': 'Connect'}
        logUrl = '/config/performLogin'
        log("%s >> logging in... " % logUrl)
        self.sleep()
        response = server.post(data, uri=logUrl)
        success = '_appy_' in server.cookies
        msg = '%s authenticated.' % config.login \
              if success else 'authentication failed.'
        log(response, 2)
        log(msg, 2)
        self.countTime(response)
        return success, response

    def runOne(self, i=0):
        '''Executes the test scenario once'''
        self.server = Resource(self.config.getUrl(i), measure=True, timeout=600)
        log = self.log
        config = self.config
        success, response = self.authenticate()
        if not success:
            log('*** Test aborted.')
            self.exit()
        # Apply the test scenario
        for hit in self.config.scenario:
            # Convert the hit to a Hit instance when relevant
            if isinstance(hit, str): hit = Hit(hit)
            success, response = hit.send(self, previous=response)
            if not success:
                self.terminate()
                log('*** Test aborted.')
                self.exit()
        self.terminate()
        log('*** Test successfully terminated.')
        log('%f second(s) waiting server responses.' % self.duration, 2,
            force=True)
        self.exit()

    def run(self):
        '''Executes the test scenario, potentially forking several child
           processes that will all execute the same scenario.'''
        config = self.config
        clients = config.clients
        if clients == 1:
            self.runOne()
        else:
            # Fork as many processes as required
            i = 0
            print('Forking %d processes...' % clients)
            while i < clients:
                pid = os.fork()
                if pid == 0:
                    # We are in the child: execute the scenario
                    self.runOne(i)
                else:
                    # We are in the parent
                    print('Child %d forked.' % pid)
                # m_runOne will exit. So here we are in the parent process.
                # Continue the loop.
                i += 1
            print('Check log in %s.' % config.logTarget)

    def terminate(self):
        '''Terminates the test by logging out of the tested server'''
        logoutUrl = '/config/performLogout'
        self.log("%s >> logging out... " % logoutUrl)
        self.server.get(logoutUrl)

    def exit(self):
        '''Exists the program'''
        return sys.exit(0)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import time

# ------------------------------------------------------------------------------
NO_LEAVE_TIME = 'I cannot compute duration of this call, the leaveTime is ' \
                'unknown.'
CALL_STACK_ERROR = "Problem with the method call stack: we did not enter any " \
                   "method, so we can't leave a method now."

# ------------------------------------------------------------------------------
class MethodCall:
    '''Stores information about a method call'''
    def __init__(self, methodName, enterTime, method=None):
        self.methodName = methodName # The name of the called method
        self.method = method # The method
        self.enterTime = enterTime # Whe did we enter the method ?
        self.leaveTime = None # When did we leave the method ?

    def duration(self):
        '''How long did the call lasted (in seconds) ?'''
        if not self.leaveTime:
            raise Profiler.Error(NO_LEAVE_TIME)
        return self.leaveTime - self.enterTime

class Profiler:
    '''Counts the time spent in some of your code's methods'''

    class Error(Exception): pass
    
    def __init__(self, obj, mustLog=True, mustPrint=False):
        self.obj = obj
        # Must I log profiling results via p_obj.log ?
        self.mustLog = obj and mustLog
        self.mustPrint = mustPrint # Must I print profiling results to stdout?
        self.init()
        self.log('profiler in use.')

    def init(self):
        '''(re-)initialize this profiler's data structures'''
        self.callStack = [] # Stack of profiled method calls
        self.durations = {} # ~{m_methodName: f_totalTimeInMethod}~
        
    def log(self, msg):
        # First, indent the message for readability
        blanks = ' ' * 2 * len(self.callStack)
        message = blanks + msg
        if self.mustLog:
            self.obj.log(message)
        if self.mustPrint:
            print(message)

    def enter(self, method):
        '''Is called when entering a given method'''
        enterTime = time.time()
        if isinstance(method, basestring):
            methodName = method
            theMethod = None
        else:
            methodName = method.__name__
            theMethod = method
        if not self.callStack:
            self.log('entering %s...' % methodName)
        self.callStack.append(MethodCall(methodName, enterTime, theMethod))
        if not self.durations.has_key(methodName):
            self.durations[methodName] = 0

    def leave(self):
        '''Is called when leaving a given method'''
        if not self.callStack:
            raise Profiler.Error(CALL_STACK_ERROR)
        methodCall = self.callStack.pop()
        methodCall.leaveTime = time.time()
        self.durations[methodCall.methodName] += methodCall.duration()
        if not self.callStack:
            self.log('leaving %s.' % methodCall.methodName)
            self.log('total durations in methods:')
            durations = self.durations.items()
            # Sort methods by total duration
            durations.sort(key=lambda x: x[1])
            for methodName, duration in durations:
                self.log('  %s: %f second(s)' % (methodName, duration))
            # Reinitialise the profiler
            self.init()
# ------------------------------------------------------------------------------



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from appy.px import Px
from appy.fields import Field

# ------------------------------------------------------------------------------
class Switch(Field):
    '''Complex field made of several sub-sets of fields among which only one is
       chosen. This field allows to have a part of a form being variable.
       The selected sub-set depends on a master/slave relationship that must be
       established between a Switch field and some other master field.
    '''
    pxView = pxEdit = pxCell = Px('''
     <x var="fieldset,fields=field.getChosenFields(zobj)" if="fields">
      <!-- Remember the chosen fieldset in this hidden input field -->
      <input if="layoutType == 'edit'" type="hidden"
             name=":field.name" value=":fieldset"/>
      <x var="fieldName=None;
              groupedFields=zobj.getGroupedFields(layoutType, field.pageName,
                              cssJs=None, fields=fields)">:obj.pxFields</x>
     </x>''')
    pxSearch = ''

    def __init__(self, fields, validator=None, show=True, page='main',
      group=None, layouts=None, move=0, specificReadPermission=False,
      specificWritePermission=False, width=None, height=None, maxChars=None,
      colspan=1, master=None, masterValue=None, focus=False, mapping=None,
      generateLabel=None, label=None, scolspan=1, swidth=None, sheight=None,
      inlineEdit=False, view=None, cell=None, edit=None, xml=None,
      translations=None):
        # p_fields must be a tuple of fieldsets of the form
        #                        ~((s_name, fields),)~
        # Within this tuple, every "fields" entry is itelf a tuple of the form
        #                         ~((s_name, Field),)~
        self.fields = fields
        # Call the base Field constructor
        Field.__init__(self, validator, (0,1), None, None, show, page, group,
          layouts, move, False, True, None, False, specificReadPermission,
          specificWritePermission, width, height, None, colspan, master,
          masterValue, focus, False, mapping, generateLabel, label, False,
          scolspan, swidth, sheight, True, inlineEdit, view, cell, edit, xml,
          translations)

    def getChosenFields(self, obj, layoutType='view', fieldset=None):
        '''Returns, among self.fields, the chosen sub-set, as a "flat" list of
           Field instances.

           More precisely, r_ is a tuple (name, fields), "name" being the name
           of the chosen fieldset and "fields" being the flat list of
           corresponding fields.
        '''
        req = obj.REQUEST
        # Determine the name of the chosen fieldset. Get it from p_fieldset or
        # from the request.
        master = self.master
        if not fieldset:
            # Determine the fieldset... 
            if master and master.valueIsInRequest(obj, req):
                # ... via the master value if present in the request
                requestValue = master.getRequestValue(obj)
                masterValue = master.getStorableValue(obj, requestValue,
                                                      complete=True)
                fieldset = self.masterValue(obj.appy(), masterValue)
            else:
                # ... via the stored value
                fieldset = self.getValue(obj)
                if not fieldset and not master:
                    # ... or via the first fieldset, considered as the default
                    # one if the switch has not master.
                    fieldset = self.fields[0][0]
        # Return an empty list of fields if we haven't a fieldset
        if not fieldset: return fieldset, ()
        # Get the list of fields corresponding to the chosen fieldset
        for name, fields in self.fields:
            if name == fieldset:
                return fieldset, [field for name, field in fields]
        return fieldset, ()

    def injectFields(self, klass):
        '''In order to be able to access the value of every switch sub-field as
           any other p_klass field, via expression

                               instance.[field name]

           this method, called at server startup, injects all switch sub-fields
           as p_klass attributes.
        '''
        # Note that Switch sub-fields are not added in p_klass.__fields__: only
        # "root" fields are in this attribute. So when Appy gets all fields of
        # some p_klass, it will not retrieve Switch sub-fields. So mechanisms
        # like checking fields showability and validation are performed on
        # "root" fields only, and, when relevant, on the chosen fieldset within
        # Switch fields. This is more performant and elegant. Moreover, if
        # Switch sub-fields were considered "root" fields, we would have to
        # determine global showability for it, that would be a mess.
        for fieldset, fields in self.fields:
            for name, field in fields:
                setattr(klass, name, field)
# ------------------------------------------------------------------------------



'''Pool of threads used by the Appy HTTP server'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from itertools import count
import os, queue, atexit, time, ctypes, threading

from appy.model.utils import Object as O

# Errors - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# THR = thread * WK = worker * SHUTD = shutdown

INITIAL_WK    = 'Initial worker'
ADDED_WK      = 'Worker added in response to lack of idle workers'
WK_KILLED     = 'Worker added for replacing killed thread %s'
WK_REPLACED   = 'Voluntary replacement for thread %s'
WORK_DONE     = ' > handled by thread %s (%i tasks queued).'
CHECK_ADD_THR = 'No idle workers for task; checking if we need to make ' \
                'more workers'
SPAWNING      = 'No idle tasks and only %d busy task(s): adding %d more workers'
NO_SPAWNING   = 'No extra workers needed (%d busy worker(s))'
CULL_WORKERS  = 'Culling %s extra workers. %s idle workers present: %s'
CULL_KILLED   = 'Killed thread %s no longer around'
KILL_KO       = 'PyThreadState_SetAsyncExc failed'
KILLING_THR   = 'Killing thread with ID=%s...'
KILLED_HUNGED = 'Workers killed forcefully'
WK_TO_KILL    = 'Killing thread "%s"... (working on task for %i seconds)'
HUNG_STATUS   = "killHungThreads status: %d threads (%d working, %d idle, " \
                "%d starting), average time %s, maxTime %.2fsec, killed %d " \
                "worker(s)."
ZOMBIES_FOUND = 'Zombie thread(s) found: %s.'
EXITING       = 'Exiting process because %s zombie threads is more than %s limit.'
NEW_WORKER    = 'Started new worker %s: %s'
STOPPING_THR  = 'Thread %s processed %i requests (limit %s); stopping it'
SHUTD_RECV    = 'Worker %s is asked to shutdown'
SHUTD_POOL    = 'Shutting down the threadpool (%d threads)'
SHUTD_HUNG    = "%s worker(s) didn't stop properly, and %s zombie(s)"
SHUTD_OK      = 'All workers stopped'
SHUTD_FORCED  = 'Forcefully exiting process'
SHUTD_F_OK    = 'All workers eventually killed'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class ThreadPool:
    '''Pool of threads for efficiently managing incoming client requests'''

    # This excellent code was initially copied from the Paste project, many many
    # thanks. https://github.com/cdent/paste/blob/master/paste/httpserver.py

    # This pool manages a list of thread workers managing cliet requests via a
    # queue of callables.

    # The pool keeps a notion of the status of its worker threads.
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # If...  | The worker thread ...
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # idle   | does nothing
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # busy   | is doing its job
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # hung   | has been doing a job for too long
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # dying  | is a hung thread that has been killed, but hasn't died quite yet
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # zombie | that we've tried to kill but isn't dead yet
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # The pool also maintains, in attribute "tracked", a dictionary with these
    # keys and lists of thread IDs that fall in that status. All keys will be
    # present, even if they point to emty lists.

    # Hung threads are threads that have been busy more than
    # config.hungThreadLimit seconds. Hung threads are killed when they live
    # longer than config.killThreadLimit seconds. A thread is then considered
    # dying for config.dyingLimit seconds, if it is still alive after that it is
    # considered a zombie.

    # When there are no idle workers and a request comes in, another worker
    # *may* be spawned. If there are less than config.spawnIfUnder threads in
    # the busy state, another thread will be spawned.  So if the limit is 5, and
    # there are 4 hung threads and 6 busy threads, no thread will be spawned.

    # When there are more than config.maxZombieThreadsBeforeDie zombie threads,
    # a SystemExit exception will be raised, stopping the server. Use 0 or None
    # to never raise this exception. Zombie threads *should* get cleaned up, but
    # killing threads is no necessarily reliable. This is turned off by default,
    # since it is only a good idea if you've deployed the server with some
    # process watching from above (something similar to daemontools or zdaemon).

    # Each worker thread only processes config.maxRequests tasks before it dies
    # and replaces itself with a new worker thread.

    SHUTDOWN = object()

    # This flag specifies if, when the server terminates, all in-progress client
    # connections should be droppped.
    daemon = False

    def __init__(self, server, config):
        '''Initialise the pool's data stuctures'''
        # A reference to the p_server
        self.server = server
        self.logger = server.loggers.app
        # The server config
        self.config = config
        # The Queue object allowing inter-thread communication
        self.queue = queue.Queue()
        # The list of Thread instances
        self.workers = []
        self.workersCount = count()
        # Dict storing info about thread's statuses
        self.tracked = {} # ~{i_threadId: (i_timeStarted, info)}~
        # IDs of threads being idle
        self.idleWorkers = []
        # Dict of threads having been killed, but maybe aren't dead yet
        self.dyingThreads = {} # ~{i_threadId: (i_time, Thread)}~
        # This is used to track when we last had to add idle workers
        # (storing time.time()); we shouldn't cull extra workers until some time
        # has passed (config.hungThreadLimit) since workers were added.
        self.lastAddedNewIdleWorkers = 0
        # Number of requests received since last check for hunged threads
        self.requestsSinceLastHungCheck = 0
        if not ThreadPool.daemon:
            atexit.register(self.shutdown)
        # Create the threads
        for i in range(config.threads):
            self.addWorker(message=INITIAL_WK)

    def debug(self, message):
        '''Logs a message'''
        self.logger.debug(message)

    def addTask(self, task):
        '''Adds a task to the queue'''
        # Kill hung threads if it's time to do it
        cfg = self.config
        if cfg.hungCheckPeriod:
            self.requestsSinceLastHungCheck += 1
            if self.requestsSinceLastHungCheck > cfg.hungCheckPeriod:
                self.requestsSinceLastHungCheck = 0
                self.killHungThreads()
        # Add workers when relevant
        if not self.idleWorkers and cfg.spawnIfUnder:
            # spawnIfUnder can come into effect. Count busy threads.
            busy = 0
            now = time.time()
            self.debug(CHECK_ADD_THR)
            for worker in self.workers:
                # Ignore uninitialized threads
                if not hasattr(worker, 'thread_id'): continue
                started, info = self.tracked.get(worker.thread_id, (None,None))
                if started is not None:
                    if (now - started) < cfg.hungThreadLimit:
                        busy += 1
            if cfg.spawnIfUnder > busy:
                delta = cfg.spawnIfUnder - busy
                self.debug(SPAWNING % (busy, delta))
                self.lastAddedNewIdleWorkers = time.time()
                for i in range(delta):
                    self.addWorker(message=ADDED_WK)
            else:
                self.debug(NO_SPAWNING % busy)
        # Kill supernumerary workers when relevant
        idle = len(self.idleWorkers)
        if (len(self.workers) > cfg.threads) and (idle > 3) and \
           ((time.time()-self.lastAddedNewIdleWorkers) > cfg.hungThreadLimit):
            # We've spawned workers in the past, but they aren't needed anymore;
            # kill off some.
            kill = len(self.workers) - cfg.threads
            self.debug(CULL_WORKERS % (kill, idle, self.idleWorkers))
            for i in range(kill):
                self.queue.put(self.SHUTDOWN)
        # Add the task to the queue
        self.queue.put(task)

    def cullThreadIfDead(self, id):
        '''Cull thread with this p_id if it does not exist anymore. Return True
           if the thread has been culled.'''
        if self.threadExists(id): return
        self.debug(CULL_KILLED % id)
        try:
            del self.dyingThreads[id]
            return True
        except KeyError:
            pass

    def getTracked(self, formatted=True):
        '''Returns a dict summarizing info about the threads in the pool'''
        r = O(idle=[], busy=[], hung=[], dying=[], zombie=[])
        now = time.time()
        cfg = self.config
        for worker in self.workers:
            # Ignore threads not being fully started up
            if not hasattr(worker, 'thread_id'): continue
            started, info = self.tracked.get(worker.thread_id, (None, None))
            if started is not None:
                attr = 'hung' if (now-started) > cfg.hungThreadLimit else 'busy'
            else:
                attr = 'idle'
            getattr(r, attr).append(worker)
        # Add dying and zombies
        for id, (killed, worker) in self.dyingThreads.items():
            culled = self.cullThreadIfDead(id)
            if culled: continue
            attr = 'zombie' if now - killed > cfg.dyingLimit else 'dying'
            getattr(r, attr).append(worker)
        if formatted:
            # Return it as a XHTML table
            rows = []
            for name, threads in r.items():
                info = []
                for thread in threads:
                    info.append('%s - %s' % (thread.getName(), thread.ident))
                rows.append('<tr><th>%s</th><td>%s</td></tr>' % \
                            (name, '<br/>'.join(info)))
            r = '<table class="small"><tr><th>Type</th><th>Threads</th></tr>' \
                '%s</table>' % '\n'.join(rows)
        return r

    def killWorker(self, id):
        '''Removes from the pool, the worker whose thread ID is p_id and
           replaces it with a new worker thread.'''
        # This should only be done for mis-behaving workers
        thread = threading._active.get(id)
        # Kill the thread by raising SystemExit
        tid = ctypes.c_long(id)
        r = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid,
              ctypes.py_object(SystemExit))
        if r == 0:
            pass # Invalid thread id: it has died in the mean time
        elif r != 1:
            # If it returns a number greater than one, you're in trouble,
            # and you should call it again with exc=NULL to revert the effect.
            ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, 0)
            raise SystemError(KILL_KO)
        # Remove any reference to the thread in the pool
        try:
            del self.tracked[id]
        except KeyError:
            pass
        self.debug(KILLING_THR % id)
        if thread in self.workers:
            self.workers.remove(thread)
        self.dyingThreads[id] = (time.time(), thread)
        self.addWorker(message=WK_KILLED % id)

    def addWorker(self, *args, **kwargs):
        index = next(self.workersCount)
        worker = threading.Thread(target=self.workerMethod, args=args,
                                  kwargs=kwargs, name=("worker %d" % index))
        worker.setDaemon(ThreadPool.daemon)
        worker.start()

    def threadExists(self, id):
        '''Returns True if a thread with this p_id exists'''
        return id in threading._active

    def killHungThreads(self):
        '''Tries to kill any hung thread'''
        cfg = self.config
        # No killing should occur in that case
        if not cfg.killThreadLimit: return
        now = time.time()
        maxTime = totalTime = idle = starting = working = killed = 0
        for worker in self.workers:
            # Ignore uninitialized threads
            if not hasattr(worker, 'thread_id'):
                starting += 1
                continue
            id = worker.thread_id
            started, info = self.tracked.get(id, (None, None))
            if started is None:
                # Must be idle
                idle += 1
                continue
            lifetime = now - started
            working += 1
            maxTime = max(maxTime, lifetime)
            totalTime += lifetime
            if lifetime > cfg.killThreadLimit:
                self.logger.warn(WK_TO_KILL % (worker.getName(), lifetime))
                self.killWorker(id)
                killed += 1
        average = '%.2fsec' % (totalTime / working) if working else 'N/A'
        self.debug(HUNG_STATUS % ((idle + starting + working), working, idle, \
                                  starting, average, maxTime, killed))
        self.checkMaxZombies()

    def checkMaxZombies(self):
        '''Checks if we've reached cfg.maxZombieThreadsBeforeDie; if so, kill
           the entire process.'''
        cfg = self.config
        if not cfg.maxZombieThreadsBeforeDie: return
        found = []
        now = time.time()
        for id, (killed, worker) in self.dyingThreads.items():
            culled = self.cullThreadIfDead(id)
            if culled: continue
            if (now - killed) > cfg.dyingLimit:
                found.append(id)
        if found:
            self.debug(ZOMBIES_FOUND % ', '.join(found))
        count = len(found)
        if count > cfg.maxZombieThreadsBeforeDie:
            self.logger.error(EXITING % (count, cfg.maxZombieThreadsBeforeDie))
            self.shutdown(10)
            raise SystemExit(3)

    def workerMethod(self, message=None):
        '''Method executed by a worker thread'''
        cfg = self.config
        thread = threading.currentThread()
        id = thread.thread_id = threading.get_ident()
        self.workers.append(thread)
        self.idleWorkers.append(id)
        processed = 0 # The number of processed requests
        replace = False # Must we replace the thread ?
        self.debug(NEW_WORKER % (id, message))
        try:
            while True:
                # Stop myself if I have managed the maximum number of requests
                max = cfg.maxRequests
                if max and (processed > max):
                    # Replace me then die
                    self.debug(STOPPING_THR % (id, processed, max))
                    replace = True
                    break
                # Get the job I have to do
                runnable = self.queue.get()
                if runnable is ThreadPool.SHUTDOWN:
                    # This is not a job, but a shutdown command
                    self.debug(SHUTD_RECV % id)
                    break
                # Note myself as working
                try:
                    self.idleWorkers.remove(id)
                except ValueError:
                    pass
                self.tracked[id] = [time.time(), None]
                processed += 1
                # Perform my task
                runnable()
                self.debug(WORK_DONE % (id, self.queue.qsize()))
                # Note myself again as being "idle"
                try:
                    del(self.tracked[id])
                except KeyError:
                    pass
                self.idleWorkers.append(id)
        finally:
            try:
                del(self.tracked[id])
            except KeyError:
                pass
            try:
                self.idleWorkers.remove(id)
            except ValueError:
                pass
            try:
                self.workers.remove(thread)
            except ValueError:
                pass
            try:
                del(self.dyingThreads[id])
            except KeyError:
                pass
            if replace:
                self.addWorker(message=WK_REPLACED % id)

    def shutdown(self, forceQuitTimeout=0):
        '''Shutdown the queue (after finishing any pending requests)'''
        count = len(self.workers)
        self.debug(SHUTD_POOL % count)
        # Add a shutdown request for every worker
        for i in range(count):
            self.queue.put(ThreadPool.SHUTDOWN)
        # Wait for each thread to terminate
        hung = []
        for worker in self.workers:
            worker.join(0.5)
            if worker.is_alive():
                hung.append(worker)
        zombies = []
        for id in self.dyingThreads:
            if self.threadExists(id):
                zombies.append(id)
        if hung or zombies:
            self.debug(SHUTD_HUNG % (len(hung), len(zombies)))
            if hung:
                for worker in hung:
                    self.killWorker(worker.thread_id)
                self.debug(KILLED_HUNGED)
            if forceQuitTimeout:
                timedOut = False
                needForceQuit = bool(zombies)
                for worker in self.workers:
                    if not timedOut and worker.is_alive():
                        timedOut = True
                        worker.join(forceQuitTimeout)
                    if worker.is_alive():
                        # Worker won't die
                        needForceQuit = True
                if needForceQuit:
                    # Remove the threading atexit callback
                    for callback in list(atexit._exithandlers):
                        func = getattr(callback[0], 'im_func', None)
                        if not func:
                            continue
                        globs = getattr(func, 'func_globals', {})
                        mod = globs.get('__name__')
                        if mod == 'threading':
                            atexit._exithandlers.remove(callback)
                    atexit._run_exitfuncs()
                    self.debug(SHUTD_FORCED)
                    os._exit(3)
                else:
                    self.debug(SHUTD_F_OK)
        else:
            self.debug(SHUTD_OK)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''job.py must be executed by a "zopectl run" command and, as single arg,
   must get a string with the following format:

    <ZopeAdmin>:<PloneInstancePath>:<ApplicationName>:<ToolMethodName>[:<args>].

     <ZopeAdmin> is the userName of the Zope administrator for this instance.
     <PloneInstancePath> is the path, within Zope, to the Plone Site object (if
                         not at the root of the Zope hierarchy, use '/' as
                         folder separator); leave blank if using appy.gen > 0.8

     <ApplicationName> is the name of the Appy application. If it begins with
                       "path=", it does not represent an Appy application, but
                       the path, within <PloneInstancePath>, to any Zope object
                       (use '/' as folder separator); leave blank if using
                       appy.gen > 0.8;

     <ToolMethodName> is the name of the method to call on the tool in this
                      Appy application, or the method to call on the arbitrary
                      Zope object if previous param starts with "path=".

     <args> (optional) are the arguments to give to this method (only strings
            are supported). Several arguments must be separated by '*'.

    Note that you can also specify several commands, separated with
    semicolons (";"). This scripts performs a single commit after all commands
    have been executed.
'''

# ------------------------------------------------------------------------------
import sys, transaction

# Check that job.py is called with the right parameters.
if len(sys.argv) != 2:
    print('job.py was called with wrong args.')
    print(__doc__)
else:
    commands = sys.argv[1].split(';')
    # Check that every command has the right number of sub-elelements.
    for command in commands:
        parts = command.split(':')
        if len(parts) not in (4,5):
            print('job.py was called with wrong args.')
            print(__doc__)

    for command in commands:
        parts = command.split(':')
        # Unwrap parameters
        if len(parts) == 4:
            zopeUser, plonePath, appName, toolMethod = parts
            args = ()
        else:
            zopeUser, plonePath, appName, toolMethod, args = parts
        # Zope was initialized in a minimal way. Complete Zope install.
        from Testing import makerequest
        app = makerequest.makerequest(app)
        app.REQUEST._fake_ = True
        # Log as Zope admin
        from AccessControl.SecurityManagement import newSecurityManager
        user = app.acl_users.getUserById(zopeUser)
        if not user:
            # Try with user "admin"
            user = app.acl_users.getUserById('admin')
        if not hasattr(user, 'aq_base'):
            user = user.__of__(app.acl_users)
        newSecurityManager(None, user)
        # Find the root object.
        rootObject = app # Initialised with the Zope root object.
        if plonePath:
            for elem in plonePath.split('/'):
                rootObject = getattr(rootObject, elem)
        # If we are in a Appy application, the object on which we will call the
        # method is the config object on this root object.
        if not appName:
            targetObject = rootObject.config.appy()
        elif not appName.startswith('path='):
            objectName = 'portal_%s' % appName.lower()
            targetObject = getattr(rootObject, objectName).appy()
        else:
            # It can be any object.
            targetObject = rootObject
            for elem in appName[5:].split('/'):
                targetObject = getattr(targetObject, elem)
        # Execute the method on the target object
        if args: args = args.split('*')
        exec 'targetObject.%s(*args)' % toolMethod
    transaction.commit()
# ------------------------------------------------------------------------------
